package com.hand.hls.web.svc;

import com.hand.hap.attachment.service.IAttachmentProvider;
import com.hand.hap.cache.Cache;
import com.hand.hap.cache.CacheManager;
import com.hand.hap.cache.impl.ResourceCustomizationCache;
import com.hand.hap.core.BaseConstants;
import com.hand.hap.core.web.FreeMarkerBeanProvider;
import com.hand.hap.core.web.view.XMap;
import com.hand.hap.core.web.view.XMapHandler;
import com.hand.hap.core.web.view.XMapParser;
import com.hand.hap.core.web.view.XMapUtil;
import com.hand.hap.function.dto.Resource;
import com.hand.hap.system.service.IAccessService;
import com.hand.hls.web.svc.components.BmOperation;
import com.hand.hls.web.svc.components.FreemarkerConfigHolder;
import freemarker.template.Configuration;
import freemarker.template.SimpleHash;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanFactoryUtils;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.context.ApplicationContextException;
import org.springframework.web.servlet.support.RequestContextUtils;
import org.springframework.web.servlet.view.freemarker.FreeMarkerConfig;
import org.springframework.web.servlet.view.freemarker.FreeMarkerView;
import org.xmlpull.v1.XmlPullParserException;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;
import javax.script.SimpleScriptContext;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.*;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-11-27
 * Time: 15:28
 */
public class ServiceView extends FreeMarkerView {
    protected final Logger logger = LoggerFactory.getLogger(ServiceView.class);

    private ServiceTagFactory serviceTagFactory;

    /*cache*/
    private CacheManager cacheManager;
    private Cache<Resource> resourceCache;
    private ResourceCustomizationCache resourceCustomizationCache;
    private IAccessService accessService;
    private IAttachmentProvider attachmentProvider;
    private FreeMarkerBeanProvider beanProvider;
    private BmOperation bmOperation;
    private ScriptEngine engine;

    public ServiceTagFactory getServiceTagFactory() {
        return serviceTagFactory;
    }

    @Override
    protected void initServletContext(ServletContext servletContext) throws BeansException {
        super.initServletContext(servletContext);
        Configuration configuration = getConfiguration();
        FreemarkerConfigHolder.setConfiguration(configuration);
        engine = new ScriptEngineManager().getEngineByName("nashorn");
        this.serviceTagFactory = (ServiceTagFactory) autoDetectServiceTagFactory(ServiceTagFactory.class);
        this.bmOperation = (BmOperation) autoDetectServiceTagFactory(BmOperation.class);
        this.cacheManager = (CacheManager) autoDetectServiceTagFactory(CacheManager.class);
        this.resourceCustomizationCache = (ResourceCustomizationCache) autoDetectServiceTagFactory(ResourceCustomizationCache.class);
        resourceCache = cacheManager.getCache(BaseConstants.CACHE_RESOURCE_URL);
    }

    private Object autoDetectServiceTagFactory(Class<?> clazz) throws BeansException {
        try {
            return BeanFactoryUtils.beanOfTypeIncludingAncestors(getApplicationContext(), clazz, true, false);
        } catch (NoSuchBeanDefinitionException ex) {
            throw new ApplicationContextException("Must define a single ServiceTagFactory bean in this web application context!", ex);
        }
    }

    private void processServiceTemplate(Template template, Map<String, Object> model, HttpServletRequest request, HttpServletResponse response) throws Exception {
        SimpleHash fmModel = buildTemplateModel(model, request, response);
        try (Writer writer = response.getWriter()) {
            XMap xmap = parseService(fmModel, template);

            Resource resource = resourceCache.getValue(template.getName());
            if (resource != null) {
                String[] mergeList = resourceCustomizationCache.getValue(resource.getResourceId().toString());
                if (mergeList != null) {
                    List<String> list = Arrays.asList(mergeList);
                    mergeService(xmap, list, request, fmModel);
                }
            }

            String str = ServiceBuilder.build(xmap, new ServiceContext(getConfiguration(), getServiceTagFactory(), fmModel.toMap(), request, response));
            if (str != null) {
                Object enableDebug = request.getAttribute("enableDebug");
                if ("Y".equals(enableDebug) && logger.isDebugEnabled()) {
                    logger.debug("parse service content:\n{}", str);
                }
                HashMap<String, Object> map = prepareRequestParams(request);
                //TODO: run js here
                // clear engine context
                HttpSession session = request.getSession();
                engine.setContext(new SimpleScriptContext());
                engine.put("$bm", bmOperation);
                engine.put("$session", session);
                engine.put("$params", map);
                engine.put("$log", logger);
                engine.put("$request", request);
                engine.eval(str);
                writer.write(str);
            }
        } catch (Exception e) {
            if (logger.isErrorEnabled()) {
                logger.error("Failed to parse service template for URL [" + getUrl() + "]", e);
            }
            throw e;
        }
    }

    private HashMap<String, Object> prepareRequestParams(HttpServletRequest request) {
        Map<String, String[]> parameterMap = request.getParameterMap();
        HashMap<String, Object> map = new HashMap<>();
        if(parameterMap != null && parameterMap.size() > 0){
            parameterMap.forEach((key,value)->{
                if(value != null){
                    if(value.length == 1){
                        map.put(key,value[0]);
                    }else{
                        map.put(key,value);
                    }
                }
            });
        }
        return map;
    }

    private XMap parseService(SimpleHash model, Template template) throws IOException, TemplateException, XmlPullParserException {
        XMap xmap;
        try (StringWriter out = new StringWriter()) {
            template.process(model, out);
            out.flush();
            xmap = XMapParser.parseReader(new StringReader(out.toString()), new XMapHandler());
        }
        return xmap;
    }

    private void mergeService(XMap source, List<String> list, HttpServletRequest request, SimpleHash model) {
        if (list != null) {
            Locale locale = RequestContextUtils.getLocale(request);
            for (String url : list) {
                Template t;
                try {
                    t = getTemplate(url, locale);
                    XMap service = parseService(model, t);
                    XMapUtil.merge(source, service);
                } catch (Exception e) {
                    if (logger.isErrorEnabled()) {
                        logger.error("Failed to merge service " + url, e);
                    }
                }

            }
        }
    }

    @Override
    protected void doRender(Map<String, Object> model, HttpServletRequest request, HttpServletResponse response) throws Exception {
        exposeModelAsRequestAttributes(model, request);
        if (logger.isDebugEnabled()) {
            logger.debug("Rendering FreeMarker template [{}] in FreeMarkerView '{}'", getUrl(), getBeanName());
        }
        Locale locale = RequestContextUtils.getLocale(request);
        // rewrite doRender to modify the template process
        processServiceTemplate(getTemplate(locale), model, request, response);
    }

    @Override
    protected FreeMarkerConfig autodetectConfiguration() throws BeansException {
        accessService = getApplicationContext().getBean(IAccessService.class);
        attachmentProvider = getApplicationContext().getBean(IAttachmentProvider.class);
        beanProvider = getApplicationContext().getBean(FreeMarkerBeanProvider.class);
        return super.autodetectConfiguration();
    }

    @Override
    protected SimpleHash buildTemplateModel(Map<String, Object> model, HttpServletRequest request,
                                            HttpServletResponse response) {
        SimpleHash fmModel = super.buildTemplateModel(model, request, response);
        accessService.setRequest(request);
        fmModel.put("accessService", accessService);
        attachmentProvider.setConfiguration(getConfiguration());
        fmModel.put("attachmentProvider", attachmentProvider);
        Map<String, Object> beans = beanProvider.getAvailableBean();
        if (beans != null) {
            fmModel.putAll(beans);
        }
        return fmModel;
    }

}