package hls.core.db.components;

import com.google.common.collect.Sets;
import org.apache.commons.lang3.StringUtils;
import org.apache.ibatis.builder.xml.XMLMapperBuilder;
import org.apache.ibatis.executor.ErrorContext;
import org.apache.ibatis.session.Configuration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.NestedIOException;
import org.springframework.core.io.Resource;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.util.*;

/**
 * mybatis Mapper刷新插件
 * <p>
 * Created by fjm on 2017/11/12.
 */
public class MapperRefreshPlugin {

    public static Logger log = LoggerFactory.getLogger(MapperRefreshPlugin.class);
    private static Properties prop = new Properties();
    private static boolean refresh;        // 刷新启用后，是否启动了刷新线程
    private Set<String> location;          // Mapper实际资源路径
    public static Resource[] mapperLocations;     // Mapper资源路径
    public static Configuration configuration;        // MyBatis配置对象
    public static Long beforeTime;           // 上一次刷新时间

    private boolean enable = false;        //默认为不启动

    public static boolean isRefresh() {
        return refresh;
    }

    public boolean isEnable() {
        return enable;
    }

    public void setEnable(boolean enable) {
        this.enable = enable;
    }

    /**
     * 初始化资源
     *
     * @param mapperLocations
     * @param configuration
     * @param beforeTime
     */
    public static void initResource(Resource[] mapperLocations, Configuration configuration, long beforeTime) {
        MapperRefreshPlugin.mapperLocations = mapperLocations;
        MapperRefreshPlugin.configuration = configuration;
        MapperRefreshPlugin.beforeTime = beforeTime;
    }

    /**
     * 开始刷新mapper
     */
    public List<File> doRefresh() {
        log.debug("[location] " + location);
        log.debug("[configuration] " + configuration);
        List<File> refreshList = null;
        if (isEnable()) {
            if (location == null) { //当location为空时,初始化location
                location = Sets.newHashSet();
                log.debug("Init MapperRefresh.location:");
                log.debug("MapperLocation's length:" + mapperLocations.length);
                for (Resource mapperLocation : mapperLocations) {
                    String s = mapperLocation.toString().replaceAll("\\\\", "/");
                    s = s.substring("file [".length());
                    s = s.substring(0, s.length() - 1);
                    if (!location.contains(s)) {
                        location.add(s);
                        log.debug("Location:" + s);
                    }
                }
                log.debug("Location's size:" + location.size());
            }
            refresh = true;
            log.info("========= Enabled refresh mybatis mapper =========");
            try {
                refreshList = refresh(location, beforeTime);
                return refreshList;
            } catch (Exception e) {
                e.printStackTrace();
            }
            log.info("========= Mapper was refreshed =========");
        }
        return refreshList;
    }

    /**
     * 执行刷新
     *
     * @param filePath   刷新的文件的集合
     * @param beforeTime 上次记录的文件更新时间
     * @throws NestedIOException     解析异常
     * @throws FileNotFoundException 文件未找到
     * @author ThinkGem
     */
    @SuppressWarnings({"rawtypes", "unchecked"})
    private List<File> refresh(Set<String> filePath, Long beforeTime) throws Exception {
        // 本次刷新时间
        Long refreshTime = System.currentTimeMillis();
        // 获取需要刷新的Mapper文件列表
        List<File> fileList = this.getRefreshFile(filePath, beforeTime);
        if (fileList.size() > 0) {
            log.debug("Refresh file: " + fileList.size());
        }
        for (int i = 0; i < fileList.size(); i++) {
            InputStream inputStream = new FileInputStream(fileList.get(i));
            String resource = fileList.get(i).getAbsolutePath();
            try {
                // 清理原有资源，更新为自己的StrictMap，方便增量重新加载
                String[] mapFieldNames = new String[]{
                        "mappedStatements", "caches",
                        "resultMaps", "parameterMaps",
                        "keyGenerators", "sqlFragments"
                };
                for (String fieldName : mapFieldNames) {
                    Field field = configuration.getClass().getDeclaredField(fieldName);
                    field.setAccessible(true);
                    Map map = ((Map) field.get(configuration));
                    if (!(map instanceof StrictMap)) {
                        Map newMap = new StrictMap(StringUtils.capitalize(fieldName) + "collection");
                        for (Object key : map.keySet()) {
                            try {
                                newMap.put(key, map.get(key));
                            } catch (IllegalArgumentException ex) {
                                newMap.put(key, ex.getMessage());
                            }
                        }
                        field.set(configuration, newMap);
                    }
                }
                // 清理已加载的资源标识，方便让它重新加载。
                Field loadedResourcesField = configuration.getClass().getDeclaredField("loadedResources");
                loadedResourcesField.setAccessible(true);
                Set loadedResourcesSet = ((Set) loadedResourcesField.get(configuration));
                loadedResourcesSet.remove(resource);
                //重新编译加载资源文件。
                XMLMapperBuilder xmlMapperBuilder = new XMLMapperBuilder(inputStream, configuration,
                        resource, configuration.getSqlFragments());
                xmlMapperBuilder.parse();
            } catch (Exception e) {
                throw new NestedIOException("Failed to parse mapping resource: '" + resource + "'", e);
            } finally {
                ErrorContext.instance().reset();
            }
            if (log.isDebugEnabled()) {
                log.debug("Refresh file: " + fileList.get(i).getAbsolutePath());
                log.debug("Refresh filename: " + fileList.get(i).getName());
            }
        }
        // 如果刷新了文件，则修改刷新时间，否则不修改
        if (fileList.size() > 0) {
            this.beforeTime = refreshTime;
        }
        return fileList;
    }

    /**
     * 获取需要刷新的文件列表
     *
     * @param filePaths  文件
     * @param beforeTime 上次刷新时间
     * @return 刷新文件列表
     */
    private List<File> getRefreshFile(Set<String> filePaths, Long beforeTime) {
        final List<File> fileList = new ArrayList<>();
        filePaths.forEach(item -> {
            File file = new File(item);
            if (this.checkFile(file, beforeTime)) {
                fileList.add(file);
            }
        });
        return fileList;
    }

    /**
     * 判断文件是否需要刷新
     *
     * @param file       文件
     * @param beforeTime 上次刷新时间
     * @return 需要刷新返回true，否则返回false
     */
    private boolean checkFile(File file, Long beforeTime) {
        if (file.lastModified() > beforeTime) {
            return true;
        }
        return false;
    }

    /**
     * 获取整数属性
     *
     * @param key
     * @return
     */
    private static int getPropInt(String key) {
        int i = 0;
        try {
            i = Integer.parseInt(getPropString(key));
        } catch (Exception e) {
        }
        return i;
    }

    /**
     * 获取字符串属性
     *
     * @param key
     * @return
     */
    private static String getPropString(String key) {
        return prop == null ? null : prop.getProperty(key);
    }

    /**
     * 重写 org.apache.ibatis.session.Configuration.StrictMap 类
     * 来自 MyBatis3.4.0版本，修改 put 方法，允许反复 put更新。
     */
    public static class StrictMap<V> extends HashMap<String, V> {

        private static final long serialVersionUID = -4950446264854982944L;
        private String name;

        public StrictMap(String name, int initialCapacity, float loadFactor) {
            super(initialCapacity, loadFactor);
            this.name = name;
        }

        public StrictMap(String name, int initialCapacity) {
            super(initialCapacity);
            this.name = name;
        }

        public StrictMap(String name) {
            super();
            this.name = name;
        }

        public StrictMap(String name, Map<String, ? extends V> m) {
            super(m);
            this.name = name;
        }

        @SuppressWarnings("unchecked")
        public V put(String key, V value) {
            // ThinkGem 如果现在状态为刷新，则刷新(先删除后添加)
            if (MapperRefreshPlugin.isRefresh()) {
                remove(key);
                MapperRefreshPlugin.log.debug("refresh key:" + key.substring(key.lastIndexOf(".") + 1));
            }
            // ThinkGem end
            if (containsKey(key)) {
                throw new IllegalArgumentException(name + " already contains value for " + key);
            }
            if (key.contains(".")) {
                final String shortKey = getShortName(key);
                if (super.get(shortKey) == null) {
                    super.put(shortKey, value);
                } else {
                    super.put(shortKey, (V) new Ambiguity(shortKey));
                }
            }
            return super.put(key, value);
        }

        public V get(Object key) {
            V value = super.get(key);
            if (value == null) {
                throw new IllegalArgumentException(name + " does not contain value for " + key);
            }
            if (value instanceof Ambiguity) {
                throw new IllegalArgumentException(((Ambiguity) value).getSubject() + " is ambiguous in " + name
                        + " (try using the full name including the namespace, or rename one of the entries)");
            }
            return value;
        }

        private String getShortName(String key) {
            final String[] keyparts = key.split("\\.");
            return keyparts[keyparts.length - 1];
        }

        protected static class Ambiguity {
            private String subject;

            public Ambiguity(String subject) {
                this.subject = subject;
            }

            public String getSubject() {
                return subject;
            }
        }
    }
}