package hls.core.db.interceptor;

import com.alibaba.druid.sql.ast.SQLStatement;
import com.alibaba.druid.sql.ast.statement.SQLDeleteStatement;
import com.alibaba.druid.sql.ast.statement.SQLUpdateStatement;
import com.alibaba.druid.sql.dialect.mysql.parser.MySqlStatementParser;
import com.alibaba.druid.sql.dialect.oracle.parser.OracleStatementParser;
import com.alibaba.druid.sql.parser.SQLStatementParser;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.xml.XmlMapper;
import com.google.common.base.CaseFormat;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.ibatis.executor.Executor;
import org.apache.ibatis.mapping.BoundSql;
import org.apache.ibatis.mapping.MappedStatement;
import org.apache.ibatis.plugin.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.jdbc.core.JdbcTemplate;

import javax.persistence.Id;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.*;

import static hls.core.db.utils.DataBackupHelper.clearNeedFilter;
import static hls.core.db.utils.DataBackupHelper.getNeedFilter;

/**
 * Created by weck on 29/01/2018 11:10 PM
 */
@Intercepts({
        @Signature(
                type = Executor.class,
                method = "update",
                args = {MappedStatement.class, Object.class})
})
public class DataBackupInterceptor implements Interceptor {

    private final static String SQL_METHOD_UPDATE = "UPDATE";
    private final static String SQL_METHOD_DELETE = "DELETE";
    private final static String DATA_VERSION = "OBJECT_VERSION_NUMBER";
    private final static String PRIMARY_KEY_OPERATION_FLAG = "primarykey";
    private final static String EN_CODE_TYPE = "utf-8";

    @Value("${data.backup.filePath}")
    private String backupFilePath;
    @Value("${data.backup}")
    private Boolean dataBackup;
    private Logger logger = LoggerFactory.getLogger(getClass());
    private String dbType;
    private JdbcTemplate jdbcTemplate;

    public JdbcTemplate getJdbcTemplate() {
        return jdbcTemplate;
    }

    public void setJdbcTemplate(JdbcTemplate jdbcTemplate) {
        this.jdbcTemplate = jdbcTemplate;
    }

    public void setDbType(String dbType) {
        this.dbType = dbType;
    }

    public String getDbType() {
        return dbType;
    }

    @Override
    public Object intercept(Invocation invocation) throws Throwable {
        if (getNeedFilter() == null) {
            if (!dataBackup) {
                return invocation.proceed();
            }
        } else if (!getNeedFilter()) {
            clearNeedFilter();
            return invocation.proceed();
        }
        clearNeedFilter();
        MappedStatement statement = (MappedStatement) invocation.getArgs()[0];
        String commandType = statement.getSqlCommandType().toString();
        /*判断是否为基于主键的update和delete*/
        if ((commandType.equals(SQL_METHOD_UPDATE) || commandType.equals(SQL_METHOD_DELETE))
                && statement.getId().toLowerCase().indexOf(PRIMARY_KEY_OPERATION_FLAG) > 0) {
            BoundSql boundSql = statement.getBoundSql(invocation.getArgs()[1]);
            String sql = boundSql.getSql();
            SQLStatementParser parser = initParser(dbType, sql);

            List<SQLStatement> statementList = parser.parseStatementList();
            if (CollectionUtils.isEmpty(statementList)) {
                return invocation.proceed();
            }
            SQLStatement sqlStatement = statementList.get(0);
            String tableName;
            /*判断SQLStatement类型，获取tableName*/
            if (sqlStatement instanceof SQLUpdateStatement) {
                tableName = ((SQLUpdateStatement) sqlStatement).getTableName().getSimpleName();
            } else if (sqlStatement instanceof SQLDeleteStatement) {
                tableName = ((SQLDeleteStatement) sqlStatement).getTableName().getSimpleName();
            } else {
                return invocation.proceed();
            }
            if (StringUtils.isEmpty(tableName)) {
                return invocation.proceed();
            }
            Map<String, Object> primaryFields = getPrimaryKeyFields(boundSql.getParameterObject());
            if (primaryFields.isEmpty()) {
                return invocation.proceed();
            }
            /*执行update之前查询记录*/
            List<Map<String, Object>> objects = jdbcTemplate.queryForList(buildQuerySQL(tableName, primaryFields));
            if (CollectionUtils.isEmpty(objects) || objects.size() > 1) {
                return invocation.proceed();
            }
            Map<String, Object> objectMap = objects.get(0);
            /*备份记录*/
            backupData(objectMap, tableName, primaryFields);
        }
        return invocation.proceed();
    }

    @Override
    public Object plugin(Object o) {
        return Plugin.wrap(o, this);
    }

    @Override
    public void setProperties(Properties properties) {

    }

    /**
     * 获取对象内部主键的键值对
     *
     * @param o 对象
     * @return 键值对 map
     */
    private Map<String, Object> getPrimaryKeyFields(Object o) {
        Field[] fields = o.getClass().getDeclaredFields();
        Map<String, Object> valueFields = new HashMap<>();
        for (Field field : fields) {
            if (field.getAnnotation(Id.class) != null) {
                field.setAccessible(true);
                try {
                    valueFields.put(field.getName(), field.get(o));
                } catch (IllegalAccessException e) {
                    logger.error("field get inner value failed,", e);
                }
                field.setAccessible(false);
            }
        }
        return valueFields;
    }

    /**
     * 根据dbType生成不同的sql解析器
     *
     * @param dbType 数据库类型
     * @param sql    sql
     * @return SQLStatementParser parser
     */
    private SQLStatementParser initParser(String dbType, String sql) {
        switch (dbType) {
            case "mysql":
                return new MySqlStatementParser(sql);
            case "oracle":
                return new OracleStatementParser(sql);
            default:
                throw new RuntimeException("could not found proper statement parser. dbType:" + String.valueOf(dbType));
        }
    }

    /**
     * 构造查询sql
     *
     * @param tableName     表名
     * @param primaryFields 主键fields
     * @return sql(String)
     */
    private String buildQuerySQL(String tableName, Map<String, Object> primaryFields) {
        StringBuilder querySql = new StringBuilder("select * from " + tableName + " where ");
        Set<String> keys = primaryFields.keySet();
        for (String key : keys) {
            querySql.append(CaseFormat.LOWER_CAMEL.to(CaseFormat.LOWER_UNDERSCORE, key))
                    .append("=")
                    .append(primaryFields.get(key)).append(" and ");
        }
        return StringUtils.removeEnd(querySql.toString(), "and ");
    }

    /**
     * 备份记录
     *
     * @param objectMap     查询出来的数据对象
     * @param tableName     表名
     * @param primaryFields 主键fields
     */
    private void backupData(Map<String, Object> objectMap, String tableName, Map<String, Object> primaryFields) {
        ObjectMapper mapper = new XmlMapper();
        try {
            String xmlString = mapper.writeValueAsString(objectMap);
            String backupDir = genFileDir(tableName);
            File backupFileDir = new File(backupDir);
            if (!backupFileDir.exists()) {
                if (!backupFileDir.mkdirs()) {
                    return;
                }
            }
            File backupFile = new File(backupDir + genFileName(primaryFields, objectMap.get(DATA_VERSION)));
            if (!backupFile.exists()) {
                if (!backupFile.createNewFile()) {
                    return;
                }
            }
            IOUtils.write(xmlString.getBytes(EN_CODE_TYPE), new FileOutputStream(backupFile));
        } catch (IOException e) {
            logger.error("transfer backup object to xml file failed,tableName is {}, object data is {},", tableName, objectMap.toString(), e);
        }

    }

    /**
     * 生成备份目录文件夹
     *
     * @param tableName 文件名
     * @return dir string
     */
    private String genFileDir(String tableName) {
        StringBuilder pathBuilder = new StringBuilder(backupFilePath);
        if (StringUtils.endsWith(backupFilePath, File.separator)) {
            pathBuilder.append(File.separator);
        }
        pathBuilder.append(tableName).append(File.separator);

        Calendar calendar = Calendar.getInstance(Locale.CHINA);
        int year = calendar.get(Calendar.YEAR);
        int month = calendar.get(Calendar.MONTH);
        int date = calendar.get(Calendar.DATE);
        pathBuilder
                .append(year).append('-')
                .append(month + 1).append('-')
                .append(date)
                .append(File.separator);

        return pathBuilder.toString();
    }

    /**
     * 生成备份文件名
     *
     * @param primaryFields       主键fields
     * @param objectVersionNumber 数据版本号
     * @return filename，string
     */
    private String genFileName(Map<String, Object> primaryFields, Object objectVersionNumber) {
        if (objectVersionNumber == null) {
            return StringUtils.split(primaryFields.toString(), "}")[0] + ",timeStamp=" + System.currentTimeMillis() + "}.xml";
        }
        return StringUtils.split(primaryFields.toString(), "}")[0] + ",objectVersionNumber=" + objectVersionNumber + "}.xml";
    }
}
