package hls.core.db.service.impl;

import com.hand.hap.generator.service.impl.DBUtil;
import com.hand.hap.mybatis.common.Mapper;
import com.hand.hap.system.dto.ResponseData;
import hls.core.db.service.IExportDatabaseService;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
import java.net.URLEncoder;
import java.sql.*;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-08-24
 * Time: 10:38
 */
@Service
@Transactional(readOnly = true)
public class ExportDatabaseServiceImpl implements IExportDatabaseService {
    public static final String ENC = "UTF-8";
    public static final String FILE_NAME = "backup.sql";
    @Autowired
    @Qualifier("sqlSessionFactory")
    SqlSessionFactory sqlSessionFactory;

    @Autowired
    private JdbcTemplate jdbcTemplate;
    @Autowired
    private Map<String, Mapper> mappers;

    private SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd hh:mm:ss");

    private Logger logger = LoggerFactory.getLogger(getClass());

    @Override
    public ResponseData exportTables(List<String> tableNames, HttpServletResponse response) {
        ResponseData responseData = new ResponseData(true);
        try (ServletOutputStream outputStream = response.getOutputStream()) {
            try (OutputStreamWriter outputStreamWriter = new OutputStreamWriter(outputStream)) {
                response.addHeader("Content-Disposition", "attachment;filename=\"" + URLEncoder.encode(FILE_NAME, ENC) + "\"");
                response.setContentType("text/plain" + ";charset=" + ENC);
                response.addHeader("Accept-Ranges", "bytes");
                exportTables(tableNames, outputStreamWriter);
            }
        } catch (IOException e) {
            logger.error(e.getMessage(), e);
            responseData.setSuccess(false);
            responseData.setMessage(e.getMessage());
        }
        return responseData;
    }

    @Override
    public ResponseData exportTablesAndDatas(List<String> tableNames, List<String> dataTables, HttpServletResponse response) {
        ResponseData responseData = new ResponseData(true);
        try (ServletOutputStream outputStream = response.getOutputStream()) {
            try (OutputStreamWriter outputStreamWriter = new OutputStreamWriter(outputStream)) {
                response.addHeader("Content-Disposition", "attachment;filename=\"" + URLEncoder.encode(FILE_NAME, ENC) + "\"");
                response.setContentType("text/x-sql" + ";charset=" + ENC);
                response.addHeader("Accept-Ranges", "bytes");
                exportTables(tableNames, outputStreamWriter);
                exportDatas(dataTables, outputStreamWriter);
            }
        } catch (IOException e) {
            logger.error(e.getMessage(), e);
            responseData.setSuccess(false);
            responseData.setMessage(e.getMessage());
        }
        return responseData;
    }

    @Override
    public ResponseData exportDatas(List<String> tableNames, HttpServletResponse response) {
        ResponseData responseData = new ResponseData(true);
        try (ServletOutputStream outputStream = response.getOutputStream()) {
            try (OutputStreamWriter outputStreamWriter = new OutputStreamWriter(outputStream)) {
                response.addHeader("Content-Disposition", "attachment;filename=\"" + URLEncoder.encode(FILE_NAME, ENC) + "\"");
                response.setContentType("text/x-sql" + ";charset=" + ENC);
                response.setHeader("Accept-Ranges", "bytes");
                exportDatas(tableNames, outputStreamWriter);
            }
        } catch (IOException e) {
            logger.error(e.getMessage(), e);
            responseData.setSuccess(false);
            responseData.setMessage(e.getMessage());
        }
        return responseData;
    }

    @Override
    public ResponseData queryTables() {
        ResponseData responseData = new ResponseData();
        try {
            List<Map<String, String>> maps = new ArrayList<>();
            SqlSession sqlSession = sqlSessionFactory.openSession();
            Connection conn = DBUtil.getConnectionBySqlSession(sqlSession);
            DatabaseMetaData dbmd = conn.getMetaData();
            String database = conn.getCatalog();
            ResultSet rs = dbmd.getTables(database, null, null, new String[]{"TABLE"});
            while (rs.next()) {
                HashMap<String, String> map = new HashMap<>();
                map.put("name", rs.getString("TABLE_NAME"));
                maps.add(map);
            }
            responseData.setRows(maps);
        } catch (SQLException e) {
            logger.error(e.getMessage(), e);
        }
        return responseData;
    }

    private void exportTables(List<String> tableNames, OutputStreamWriter outputStreamWriter) throws IOException {
        if (tableNames == null || tableNames.size() < 1 || outputStreamWriter == null) {
            throw new IllegalArgumentException("tableNames or outputStream can't be null.");
        }
        for (String tableName : tableNames) {
            jdbcTemplate.query("show create table `" + tableName + "`", rowSet -> {
                String ddl = rowSet.getString("Create Table");
                try {
                    IOUtils.write(ddl, outputStreamWriter);
                    IOUtils.write(";\n", outputStreamWriter);
                } catch (IOException e) {
                    logger.error(e.getMessage(), e);
                }
            });
        }
    }

    private void exportDatas(List<String> tableNames, OutputStreamWriter outputStreamWriter) throws IOException {
        if (tableNames == null || tableNames.size() < 1 || outputStreamWriter == null) {
            throw new IllegalArgumentException("tableNames or outputStream can't be null.");
        }
        for (String tableName : tableNames) {
            jdbcTemplate.query("select * from `" + tableName + "`", rowSet -> {
                StringBuilder sb = new StringBuilder("insert into `").append(tableName).append("` (");
                StringBuilder columns = new StringBuilder();
                StringBuilder values = new StringBuilder();
                ResultSetMetaData metaData = rowSet.getMetaData();
                int columnCount = metaData.getColumnCount();
                for (int i = 0; i < columnCount; i++) {
                    String columnName = metaData.getColumnName(i + 1);
                    int columnType = metaData.getColumnType(i + 1);
                    Object object = rowSet.getObject(columnName);
                    String value = getValue(object, columnType);
                    if(value == null){
                        continue;
                    }
                    columns.append("`").append(columnName).append("`");
                    values.append(value);
                    if (i < columnCount - 1) {
                        values.append(",");
                        columns.append(",");
                    }
                }
                if(columns.charAt(columns.length()-1) == ','){
                    columns.deleteCharAt(columns.length() - 1);
                }
                if(values.charAt(values.length()-1) == ','){
                    values.deleteCharAt(values.length() - 1);
                }
                sb.append(columns).append(") values (").append(values).append(");\n");
                try {
                    IOUtils.write(sb.toString(), outputStreamWriter);
                } catch (IOException e) {
                    logger.error(e.getMessage(), e);
                }
            });
        }
    }

    private String getValue(Object object, int columnType) {
        if (object == null) {
            return null;
        }
        String value = String.valueOf(object);
        JDBCType jdbcType = JDBCType.valueOf(columnType);
        if (jdbcType != null) {
            switch (jdbcType) {
                //数字
                case NUMERIC:
                case BIT:
                case TINYINT:
                case SMALLINT:
                case INTEGER:
                case BIGINT:
                case FLOAT:
                case REAL:
                case DOUBLE:
                case DECIMAL:
                    break;
                //日期
                case DATE:
                case TIME:
                case TIMESTAMP:
                    //pass其他，包含字符串和其他特殊类型
                default:
                    value = "'" + value + "'";
            }
        } else if (Number.class.isAssignableFrom(object.getClass())) {
            if (object instanceof Double) {
                BigDecimal bigDecimal = new BigDecimal(value);
                value = bigDecimal.toPlainString();
            }
        } else {
            value = "'" + value + "'";
        }
        return value;
    }
}
