package hls.core.document.components;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.hand.hap.core.IRequest;
import com.hand.hap.core.impl.RequestHelper;
import com.hand.hap.function.dto.Resource;
import com.hand.hap.function.mapper.ResourceMapper;
import com.hand.hap.lock.components.DatabaseLockProvider;
import com.hand.hap.mybatis.entity.Example;
import com.hand.hap.system.service.IBaseService;
import hls.core.csh.service.impl.CshConContractServiceImpl;
import hls.core.document.exception.DocumentStatusForbidenException;
import hls.core.document.service.IDocumentService;
import hls.core.document.service.IDocumentStatusCheckService;
import hls.core.document.utils.DocumentCategories;
import hls.core.document.utils.DocumentUtil;
import hls.core.hls.components.DocumentChangeCommon;
import hls.core.hls.dto.DocumentStatus;
import hls.core.hls.mapper.DocumentCategoryMapper;
import hls.core.hls.mapper.DocumentStatusMapper;
import hls.core.sys.dto.SysModule;
import hls.core.sys.mapper.SysModuleMapper;
import org.apache.commons.beanutils.BeanUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.aop.support.AopUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import javax.annotation.PostConstruct;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 获取单据url
 * <p>
 * Created by fjm on 2018/1/17.
 */
@Component
@Transactional(rollbackFor = Exception.class)
public class EntranceComponent {

    private final static Logger logger = LoggerFactory.getLogger(EntranceComponent.class);

    private static final String SUFFIX = ".view?";
    public static final String DOCUMENT_CATEGORY = "documentCategory";

    private static Map<String, IBaseService> documentServiceMap = new HashMap<>();
    private static Map<String, Class> tableClassMap = new HashMap<>();

    @Autowired
    private DatabaseLockProvider databaseLockProvider;
    @Autowired
    private DocumentCategoryMapper documentCategoryMapper;
    @Autowired
    private DocumentStatusMapper documentStatusMapper;
    @Autowired
    private SysModuleMapper sysModuleMapper;
    @Autowired
    private ResourceMapper resourceMapper;
    @Autowired
    private Map<String, IBaseService> serviceMap;
    @Autowired(required = false)
    private IDocumentService documentService;
    @Autowired(required = false)
    private IDocumentStatusCheckService documentStatusCheckService;

    @PostConstruct
    public void init() {
        //系统中定义的documentCategory的集合
        List<String> documentCategoryCollect = documentCategoryMapper.selectAll().stream().map(item -> item.getDocumentCategory()).collect(Collectors.toList());
        serviceMap.forEach((k, v) -> {
            Class serviceClazz = null;
            if (!AopUtils.isAopProxy(v)) {//判断是否为代理对象
                serviceClazz = v.getClass();
            } else {
                serviceClazz = AopUtils.getTargetClass(v);
            }
            Type genType = serviceClazz.getGenericSuperclass();
            if (serviceClazz.isAssignableFrom(CshConContractServiceImpl.class)) {
                return;
            }
            if (genType instanceof ParameterizedType) {
                Type[] params = ((ParameterizedType) genType).getActualTypeArguments();
                if (params.length > 0) {
                    String tableName = DocumentChangeCommon.getTableName((Class) params[0]);
                    if (documentCategoryCollect.contains(tableName.toUpperCase()) || documentCategoryCollect.contains(tableName.toLowerCase())) {
                        documentServiceMap.put(tableName, v);
                        tableClassMap.put(tableName, (Class) params[0]);
                    }
                }
            }
        });
        logger.info("documentMap init complete");
    }

    public String getEntranceUrlByDocument(Map<String, Object> parameterMap) {
        Long documentId = null;
        String documentCategory = null;
        String functionUsage = null;
        if (parameterMap.get("documentId") != null) {
            documentId = Long.parseLong(parameterMap.get("documentId").toString());
        }
        if (parameterMap.get("documentCategory") != null) {
            documentCategory = parameterMap.get("documentCategory").toString().toUpperCase();
        }
        if (parameterMap.get("functionUsage") != null) {
            functionUsage = parameterMap.get("functionUsage").toString();
        }
        String urlStartPart = getUrlStartPart(documentCategory);//exp:start:/con/contract_
        String urlMiddlePart = getUrlMiddlePart(documentCategory, documentId, functionUsage);//exp:mid:create.view?
        String urdEndPart = getUrlEndPart(documentCategory, documentId);//exp:end:contractId=xxx
        return urlStartPart + urlMiddlePart + urdEndPart;
    }

    public String getEntranceUrlByModule(Map<String, Object> parameterMap) {
        String moduleCode = (String) parameterMap.get("menu_code");
        String code = (String) parameterMap.get("code");
        String queryParam = null;
        if (moduleCode != null) {
            queryParam = moduleCode;
        }
        if (queryParam == null && code != null) {
            queryParam = code;
        }
        SysModule sysModule = sysModuleMapper.selectByPrimaryKey(queryParam);
        Long homepageResourceId = sysModule.getHomepageResourceId();
        Resource resource = resourceMapper.selectByPrimaryKey(homepageResourceId);
        String url = resource.getUrl();
        StringBuilder result = new StringBuilder(url + "?");
        DocumentUtil.removeNullParameterAndSort(parameterMap);
        parameterMap.forEach((k, v) -> {
            result.append(k + "=" + v + "&");
        });
        if (result.toString().endsWith("&")) {
            result.delete(result.length() - 1, result.length());
        }
        if (!result.toString().startsWith("/")) {
            result.insert(0, "/");
        }
        url = result.toString();
        return url;
    }


    /**
     * 根据documentCategory获取url前缀
     *
     * @param documentCategory
     * @return
     */
    private String getUrlStartPart(String documentCategory) {
        String result = "";
        if (documentCategory == null)
            return result;
        switch (documentCategory.toUpperCase()) {
            case DocumentCategories.CON_CONTRACT:
                result = "/con/contract_";//对应合同
                break;
            case DocumentCategories.HLS_BP_MASTER:
                result = "/bp/bpmaster_";//对应商业伙伴
                break;
            case DocumentCategories.HLS_BP_MASTER_NP:
                result = "/bp/bpmaster_np_";//对应商业伙伴自然人
                break;
            case DocumentCategories.PRJ_PROJECT:
                result = "/prj/project_";//对应项目
                break;
            case DocumentCategories.LON_CONTRACT:
                result = "/lon/contract_";//对应融资合同
                break;
            case DocumentCategories.FCT_CONTRACT:
                result = "/fct/contract_";//对应保理合同
                break;
            default:
                break;
        }
        //如果存在接口的实现类,并且接口方法返回不为空,则直接使用接口返回的结果
        if (documentService != null && !StringUtils.isEmpty(documentService.getCustomerStartUrlPart(documentCategory))) {
            return documentService.getCustomerStartUrlPart(documentCategory);
        }
        return result;
    }

    /**
     * 获取url中间的部分
     *
     * @param documentCategory
     * @param documentId
     * @param functionUsage
     * @return
     */
    private String getUrlMiddlePart(String documentCategory, Long documentId, String functionUsage) {
        String result = "";
        Object queryResult = null;
        if (documentCategory == null)
            return result;
        //如果传入的id为空且documentCategory和functionUsage不为空,直接返回创建页面
        if (documentId == null && "create".equalsIgnoreCase(functionUsage)) {
            return "create" + SUFFIX;
        }
        //获取对应documentCategory对应的service对象
        IBaseService iBaseService = documentServiceMap.get(documentCategory);
        IRequest currentRequest = RequestHelper.getCurrentRequest();
        //获取对应documentCategory对应的class
        Class clazz = tableClassMap.get(documentCategory);
        try {
            Object o = clazz.newInstance();
            //获取id字段
            String idFiled = DocumentChangeCommon.getIdFliedByAnnotations(clazz);
            //将documentId设置到id字段上
            BeanUtils.setProperty(o, idFiled, documentId);
            //加锁
            databaseLockProvider.lock(o);
            //查询出对应的结果
            queryResult = iBaseService.selectByPrimaryKey(currentRequest, o);
            if (queryResult == null)
                return "create" + SUFFIX;
        } catch (InstantiationException e) {
            logger.error("InstantiationException:" + e.getMessage());
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            logger.error("IllegalAccessException:" + e.getMessage());
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            logger.error("InvocationTargetException:" + e.getMessage());
            e.printStackTrace();
        }
        result = getStatusByDocumentCategory(documentCategory, queryResult, functionUsage);
        //如果存在接口的实现类,并且接口方法返回不为空,则直接使用接口返回的结果
        if (documentService != null && !StringUtils.isEmpty(documentService.getCustomerMiddleUrlPart(documentCategory, documentId, functionUsage))) {
            return documentService.getCustomerMiddleUrlPart(documentCategory, documentId, functionUsage);
        }
        return result;
    }

    /**
     * 根据单据类型和单据id拼接参数 exp:contractId=1023&
     *
     * @param documentCategory
     * @param documentId
     * @return
     */
    private String getUrlEndPart(String documentCategory, Long documentId) {
        if (documentId == null) {
            return "";
        }
        Class clazz = tableClassMap.get(documentCategory);
        if (clazz == null) {
            return "";
        }
        String idName = DocumentChangeCommon.getIdFliedByAnnotations(clazz);
        String result = idName + "=" + documentId + "&";
        //如果存在接口的实现类,并且接口方法返回不为空,则直接使用接口返回的结果
        if (documentService != null && !StringUtils.isEmpty(documentService.getCustomerEndUrlPart(documentCategory, documentId))) {
            return documentService.getCustomerEndUrlPart(documentCategory, documentId);
        }
        return result;
    }

    /**
     * 根据documentCategory获取对象上的状态值
     *
     * @param documentCategory
     * @param o
     * @return
     */
    private static String getStatusByDocumentCategory(String documentCategory, Object o, String functionUsage) {
        String status = null;
        String json = JSON.toJSONString(o);
        Map<String, Object> map = JSONObject.parseObject(json, Map.class);
        if (DocumentCategories.CON_CONTRACT.equalsIgnoreCase(documentCategory)) {
            status = (String) map.get("contractStatus");
        } else if (DocumentCategories.LON_CONTRACT.equalsIgnoreCase(documentCategory)) {
            status = (String) map.get("contractStatus");
        } else if (DocumentCategories.PRJ_PROJECT.equalsIgnoreCase(documentCategory)) {
            status = (String) map.get("projectStatus");
        } else if (DocumentCategories.FCT_CONTRACT.equalsIgnoreCase(documentCategory)) {
            status = (String) map.get("contractStatus");
        } else if (DocumentCategories.HLS_BP_MASTER.equalsIgnoreCase(documentCategory)) {
            status = "new";
        } else if (DocumentCategories.HLS_BP_MASTER_NP.equalsIgnoreCase(documentCategory)) {
            status = "new";
        }

        if (status == null)
            status = functionUsage;
        if (status == null)
            status = "update";

        if ("update".equalsIgnoreCase(status) || "new".equalsIgnoreCase(status) || "reject".equalsIgnoreCase(status) || "cancel".equalsIgnoreCase(status) || "APPROVED_RETURN".equalsIgnoreCase(status)) {
            return "update" + SUFFIX;
        } else if ("approved".equalsIgnoreCase(status) || "approving".equalsIgnoreCase(status) || "pending".equalsIgnoreCase(status) || "change_req".equalsIgnoreCase(status)) {
            return "readonly" + SUFFIX;
        }
        return status;
    }


    /**
     * 仅用于单据提交时,校验单据状态
     *
     * @param documentId
     * @param documentCategory
     * @return
     */
    public Boolean checkDocumentStatus(Long documentId, String documentCategory) throws DocumentStatusForbidenException {

        if (documentStatusCheckService != null)
            return documentStatusCheckService.checkCustomerDocumentStatus(documentId,documentCategory);

        Class clazz = tableClassMap.get(documentCategory);//先获取documentCategory对应的类名
        if (clazz == null) {
            throw new DocumentStatusForbidenException(documentCategory, "没有找到" + documentCategory + "对应的dto,请确认dto上是否标标注表名");
        }
        IBaseService iBaseService = documentServiceMap.get(documentCategory);//获取对应的service用于查询
        if (iBaseService == null) {
            throw new DocumentStatusForbidenException(documentCategory, "没有找到" + documentCategory + "对应的service,请确认service上的泛型是否正确");
        }
        String idField = DocumentChangeCommon.getIdFliedByAnnotations(clazz);//获取id字段
        if (StringUtils.isEmpty(idField)) {
            throw new DocumentStatusForbidenException(documentCategory, "没有找到" + documentCategory + "dto上id字段,请确认是否对应字段是否包含id注解");
        }
        Object o;
        try {
            o = clazz.newInstance();
        } catch (InstantiationException e) {
            logger.error("类型{0}无法实例化,请核对{1}是否正确", clazz.getName(), documentCategory);
            e.printStackTrace();
            return false;
        } catch (IllegalAccessException e) {
            logger.error("类型{0}无法实例化,请核对{1}是否正确", clazz.getName(), documentCategory);
            e.printStackTrace();
            return false;
        }
        try {
            BeanUtils.setProperty(o, idField, documentId);//使用BeanUtils对id字段赋值
        } catch (IllegalAccessException e) {
            e.printStackTrace();
            return false;
        } catch (InvocationTargetException e) {
            e.printStackTrace();
            return false;
        }
        Object result = iBaseService.selectByPrimaryKey(RequestHelper.getCurrentRequest(), o);//拿到该单据目前的记录
        Example example = new Example(DocumentStatus.class);
        example.createCriteria().andEqualTo(DOCUMENT_CATEGORY, documentCategory);
        List<DocumentStatus> documentStatuses = documentStatusMapper.selectByExample(example);
        List<String> collect = documentStatuses.stream().map(item -> item.getDocumentCategory()).collect(Collectors.toList());
        if (DocumentCategories.PRJ_PROJECT.equalsIgnoreCase(documentCategory)) {
            String projectStatus = getStatusByBeanUtils(result, "projectStatus");
            if (!collect.contains(projectStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (projectStatus.equalsIgnoreCase("APPROVED_RETURN") || projectStatus.equalsIgnoreCase("NEW") || projectStatus.equalsIgnoreCase("CANCEL") || projectStatus.equalsIgnoreCase("REJECT")) {
                return true;
            } else if (projectStatus.equalsIgnoreCase("APPROVED") || projectStatus.equalsIgnoreCase("APPROVING")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        } else if (DocumentCategories.CON_CONTRACT.equalsIgnoreCase(documentCategory)) {
            String contractStatus = getStatusByBeanUtils(result, "contractStatus");
            if (!collect.contains(contractStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (contractStatus.equalsIgnoreCase("NEW") || contractStatus.equalsIgnoreCase("CANCEL") || contractStatus.equalsIgnoreCase("REJECT")) {
                return true;
            } else if (contractStatus.equalsIgnoreCase("BLOCK") || contractStatus.equalsIgnoreCase("ET") || contractStatus.equalsIgnoreCase("INCEPT") || contractStatus.equalsIgnoreCase("PENDING")
                    || contractStatus.equalsIgnoreCase("SIGN") || contractStatus.equalsIgnoreCase("TERMINATE")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        } else if (DocumentCategories.LON_CONTRACT.equalsIgnoreCase(documentCategory)) {
            String contractStatus = getStatusByBeanUtils(result, "contractStatus");
            if (!collect.contains(contractStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (contractStatus.equalsIgnoreCase("NEW") || contractStatus.equalsIgnoreCase("REJECT")) {
                return true;
            } else if (contractStatus.equalsIgnoreCase("APPROVING") || contractStatus.equalsIgnoreCase("APPROVED")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        } else if (DocumentCategories.LON_CONTRACT_WITHDRAW.equalsIgnoreCase(documentCategory)) {
            String withdrawStatus = getStatusByBeanUtils(result, "withdrawStatus");
            if (!collect.contains(withdrawStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (withdrawStatus.equalsIgnoreCase("NEW") || withdrawStatus.equalsIgnoreCase("REJECT")) {
                return true;
            } else if (withdrawStatus.equalsIgnoreCase("APPROVING") || withdrawStatus.equalsIgnoreCase("APPROVED")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        } else if (DocumentCategories.CON_CONTRACT_CHANGE_REQ.equalsIgnoreCase(documentCategory)) {
            String changeReqStatus = getStatusByBeanUtils(result, "status");
            if (!collect.contains(changeReqStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (changeReqStatus.equalsIgnoreCase("NEW") || changeReqStatus.equalsIgnoreCase("REJECT") || changeReqStatus.equalsIgnoreCase("CANCEL")) {
                return true;
            } else if (changeReqStatus.equalsIgnoreCase("APPROVING") || changeReqStatus.equalsIgnoreCase("APPROVED")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        } else if (DocumentCategories.CON_DEBT_EXEMPTION_REQ.equalsIgnoreCase(documentCategory)) {
            String debtExemptionReqStatus = getStatusByBeanUtils(result, "reqStatus");
            if (!collect.contains(debtExemptionReqStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (debtExemptionReqStatus.equalsIgnoreCase("NEW") || debtExemptionReqStatus.equalsIgnoreCase("REJECT") || debtExemptionReqStatus.equalsIgnoreCase("CANCEL")) {
                return true;
            } else if (debtExemptionReqStatus.equalsIgnoreCase("APPROVING") || debtExemptionReqStatus.equalsIgnoreCase("APPROVED")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        } else if (DocumentCategories.CSH_PAYMENT_REQ.equalsIgnoreCase(documentCategory)) {
            String paymentStatus = getStatusByBeanUtils(result, "status");
            if (!collect.contains(paymentStatus)) {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
            if (paymentStatus.equalsIgnoreCase("NEW") || paymentStatus.equalsIgnoreCase("REJECT") || paymentStatus.equalsIgnoreCase("CANCEL")) {
                return true;
            } else if (paymentStatus.equalsIgnoreCase("APPROVING") || paymentStatus.equalsIgnoreCase("APPROVED")) {
                return false;
            } else {
                throw new DocumentStatusForbidenException(documentCategory, "当前单据状态不合法");
            }
        }
        return false;
    }

    /**
     * 使用BeanUtils获取单据状态
     *
     * @param o
     * @param fieldName
     * @return
     */
    private static String getStatusByBeanUtils(Object o, String fieldName) {
        try {
            return BeanUtils.getProperty(o, fieldName);
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        }
        return null;
    }
}
