package hls.core.hls.components;

import com.hand.hap.core.IRequest;
import com.hand.hap.system.dto.BaseDTO;
import com.hand.hap.system.service.IBaseService;
import hls.core.cont.dto.HlsCusConContractChangeReq;
import hls.core.cont.dto.HlsCusConFloatingRateReqLn;
import hls.core.csh.dto.CshConContract;
import hls.core.csh.service.impl.CshConContractServiceImpl;
import hls.core.hls.dto.HlsCusHlsDocumentVersionRel;
import hls.core.hls.dto.HlsDocumentVersionRel;
import hls.core.hls.mapper.HlsDocumentVersionRelMapper;
import hls.core.hls.service.IHlsDocumentVersionRelService;
import hls.core.prj.dto.HlsCusHlsDocumentVersion;
import hls.core.prj.mapper.HlsCusHlsDocumentVersionMapper;
import hls.core.prj.service.HlsDocumentVersionService;
import org.apache.commons.beanutils.BeanUtils;
import org.apache.commons.beanutils.ConvertUtils;
import org.apache.commons.beanutils.converters.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.aop.support.AopUtils;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.stereotype.Component;
import org.springframework.util.StringUtils;

import javax.persistence.Table;
import javax.persistence.Id;
import javax.persistence.Transient;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.math.BigDecimal;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * 变更通用接口
 * <p>
 * 需要新增变更类型时,可以自行实现TransferToDivision接口,并定义成变量
 * <p>
 * 可参照private TransferToDivision<HlsCusConContractChangeReq> contractChange
 * <p>
 * Created by fjm on 2017/10/19.
 */
@Component
public class DocumentChangeCommon implements ApplicationContextAware {

    private final Logger logger = LoggerFactory.getLogger(this.getClass());

    public static final char UNDERLINE = '_';
    public static final String HISTORY = "HISTORY";
    public static final String CHANGE_REQ = "CHANGE_REQ";
    public static final String FIX_COLUMN_ID = "FIX_COLUMN_ID";
    public static final String FIX_COLUMN_NAME = "FIX_COLUMN_NAME";
    public static final String CON_CONTRACT_CASHFLOW = "con_contract_cashflow";

    private static Map<String, IBaseService> tableServiceMap = new HashMap<>();
    private static Map<String, Class> tableClassMap = new HashMap<>();
    private static Map<String, TransferToDivision> transferToDivisionMap = new HashMap<>();

    @Autowired
    private HlsDocumentVersionService hlsDocumentVersionService;
    @Autowired
    private HlsCusHlsDocumentVersionMapper hlsDocumentVersionMapper;
    @Autowired
    private IHlsDocumentVersionRelService hlsDocumentVersionRelService;
    @Autowired
    private HlsDocumentVersionRelMapper hlsDocumentVersionRelMapper;

    private TransferToDivision<HlsCusConContractChangeReq> contractChange = (division, changeReq) -> {
        division.setVersionDate(new Date());
        division.setSourceUserId(changeReq.getChangeReqUserId());
    };

    private TransferToDivision<HlsCusConFloatingRateReqLn> floatingChange = (division, changeReq) -> {
        division.setVersionDate(new Date());
        division.setSourceUserId(changeReq.getReqUserId());
    };

    /**
     * 初始化代码块,BeanUtils注册自定义转换器
     */
    static {
        ConvertUtils.register(new LongConverter(null), Long.class);
        ConvertUtils.register(new ShortConverter(null), Short.class);
        ConvertUtils.register(new IntegerConverter(null), Integer.class);
        ConvertUtils.register(new DoubleConverter(null), Double.class);
        ConvertUtils.register(new BigDecimalConverter(null), BigDecimal.class);
        ConvertUtils.register(new DateConverter(null), Date.class);
    }


    /**
     * 变更申请前
     *
     * @param request
     * @param dto
     */
    public HlsCusHlsDocumentVersion beforeChangeRequest(IRequest request, BaseDTO dto) {
        return changeRequestMethod(request, dto, CHANGE_REQ);
    }

    /**
     * 变更通过后
     *
     * @param dto
     * @param request
     */
    public HlsCusHlsDocumentVersion afterChangeRequest(IRequest request, BaseDTO dto) {
        return changeRequestMethod(request, dto, HISTORY);
    }

    /**
     * 根据不同的变更类型,进行变更处理
     *
     * @param request
     * @param dto
     * @param reqType
     * @return
     */
    private HlsCusHlsDocumentVersion changeRequestMethod(IRequest request, BaseDTO dto, String reqType) {
        HlsCusHlsDocumentVersion documentVersion = null;
        Class<? extends BaseDTO> clazz = dto.getClass();
        String tableName = getTableName(clazz);
        if (!StringUtils.isEmpty(tableName)) {
            TransferToDivision transferToDivision = transferToDivisionMap.get(tableName);
            if (transferToDivision == null)
                logger.info("table:{} can not match the instance of TransferToDivision , use default transition rule", tableName);
            documentVersion = getDocumentVersion(request, dto, reqType, transferToDivision);//插入一条hls_document_version
        }
        return documentVersion;
    }

    /**
     * 根据传入的dto,调用相应的service进行插入操作
     * map中的key为需要额外变更的参数名和参数值,不需要添加额外的参数请传空值
     *
     * @param request
     * @param originDto
     * @param param
     */
    public <T extends BaseDTO> T copyHistoryCommon(IRequest request, BaseDTO originDto, Map param, Long documentVersionId) {
        BaseDTO historyDto = getHistoryDto(originDto);
        Class<? extends BaseDTO> historyDtoClass = historyDto.getClass();
        Class<? extends BaseDTO> originDtoClass = originDto.getClass();
        String hisIdFliedByAnnotations = getIdFliedByAnnotations(historyDtoClass);
        String oriIdFliedByAnnotations = getIdFliedByAnnotations(originDtoClass);
        String hisTableName = getTableName(historyDtoClass);
        String oriTableName = getTableName(originDtoClass);
        if (param != null) {
            String fixColumnName = (String) param.get(FIX_COLUMN_NAME);
            Long fixColumnId = (Long) param.get(FIX_COLUMN_ID);
            try {
                BeanUtils.setProperty(historyDto, fixColumnName, fixColumnId);
                BeanUtils.setProperty(historyDto, hisIdFliedByAnnotations, null);
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            }
        }
        IBaseService iBaseService = tableServiceMap.get(hisTableName);
        if (iBaseService == null) {
            logger.warn("table:{} can not match the opposite IBaseService,please inject.", hisTableName);
        } else {
            logger.info("table:{} is already match the opposite IBaseService.", hisTableName);
            iBaseService.insertSelective(request, historyDto);
        }
        HlsCusHlsDocumentVersionRel hlsDocumentVersionRel = new HlsCusHlsDocumentVersionRel();
        hlsDocumentVersionRel.setVersionId(documentVersionId);
        hlsDocumentVersionRel.setFromTable(oriTableName);
        hlsDocumentVersionRel.setToTable(hisTableName);
        try {
            hlsDocumentVersionRel.setFromPk(Long.valueOf(BeanUtils.getProperty(originDto, oriIdFliedByAnnotations)));
            hlsDocumentVersionRel.setToPk(Long.valueOf(BeanUtils.getProperty(historyDto, hisIdFliedByAnnotations)));
            hlsDocumentVersionRelService.insertSelective(request, hlsDocumentVersionRel);
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        }
        return (T) historyDto;
    }

    /**
     * 使用history覆盖原始数据
     *
     * @param request
     * @param originDto
     * @param historyDto
     * @param param
     */
    public <T extends BaseDTO> T recoverHistoryCommon(IRequest request, T originDto, BaseDTO historyDto, Map param) {
        recoverHistoryDto(originDto, historyDto);
        Class<? extends BaseDTO> originDtoClass = originDto.getClass();
        String oriTableName = getTableName(originDtoClass);
        String idFliedByAnnotations = getIdFliedByAnnotations(originDtoClass);
        if (param != null) {
            String fixColumnName = (String) param.get(FIX_COLUMN_NAME);
            Long fixColumnId = (Long) param.get(FIX_COLUMN_ID);
            try {
                if (CON_CONTRACT_CASHFLOW.equalsIgnoreCase(oriTableName)) {
                    BeanUtils.setProperty(originDto, camel2Underline(fixColumnName), fixColumnId);
                } else {
                    BeanUtils.setProperty(originDto, fixColumnName, fixColumnId);
                }
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            }
        }
        IBaseService iBaseService = tableServiceMap.get(oriTableName);
        if (iBaseService == null) {
            logger.warn("table:{} can not match the opposite IBaseService,please inject.", oriTableName);
        } else {
            logger.info("table:{} is already match the opposite IBaseService.", oriTableName);
            try {
                String property = BeanUtils.getProperty(originDto, idFliedByAnnotations);
                if (property != null) {
                    iBaseService.updateByPrimaryKeySelective(request, originDto);
                } else {
                    iBaseService.insertSelective(request, originDto);
                }
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            } catch (NoSuchMethodException e) {
                e.printStackTrace();
            }
        }
        return originDto;
    }

    /**
     * 用change req对应的version_id，从history数据中，覆盖回原合同表(整套)
     *
     * @param request
     * @param changeReqId
     * @Deprecated
     */
    @Deprecated
    public void recoverAllHistoryByVersion(IRequest request, Long changeReqId) {
        HlsCusHlsDocumentVersion documentVersion = new HlsCusHlsDocumentVersion();
        documentVersion.setSourceDocumentId(changeReqId);
        documentVersion.setSourceDocumentCategory(CHANGE_REQ);
        documentVersion = hlsDocumentVersionMapper.selectOne(documentVersion);

        HlsDocumentVersionRel documentVersionRel = new HlsDocumentVersionRel();
        documentVersionRel.setVersionId(documentVersion.getVersionId());
        List<HlsDocumentVersionRel> hlsDocumentVersionRels = hlsDocumentVersionRelMapper.select(documentVersionRel);

        hlsDocumentVersionRels.forEach(item -> {
            String fromTable = item.getFromTable();
            Long fromPk = item.getFromPk();
            String toTable = item.getToTable();
            Long toPk = item.getToPk();
            IBaseService fromService = tableServiceMap.get(fromTable);
            IBaseService toService = tableServiceMap.get(toTable);
            BaseDTO toDto = (BaseDTO) toService.selectByPrimaryKey(request, toPk);
            BaseDTO fromDto = (BaseDTO) fromService.selectByPrimaryKey(request, fromPk);
            recoverHistoryDto(fromDto, toDto);
            fromService.updateByPrimaryKeySelective(request, fromDto);
        });

    }


    /**
     * 获取插入documentVersion记录并获取DocumentVersion对象
     *
     * @param request
     * @param changeReq
     * @param req
     * @param transferToDivision
     * @return
     */
    private HlsCusHlsDocumentVersion getDocumentVersion(IRequest request, BaseDTO changeReq, String req, TransferToDivision transferToDivision) {
        HlsCusHlsDocumentVersion hlsDocumentVersion = new HlsCusHlsDocumentVersion();
        Long changeReqId = null;
        try {
            changeReqId = Long.valueOf(BeanUtils.getProperty(changeReq, getIdFliedByAnnotations(changeReq.getClass())));
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        }
        hlsDocumentVersion.setVersionDate(new Date());
        hlsDocumentVersion.setSourceDocumentId(changeReqId);
        hlsDocumentVersion.setVersionType(req);
        hlsDocumentVersion.setSourceDocumentCategory(getTableName(changeReq.getClass()));
        if (transferToDivision != null)
            transferToDivision.transform(hlsDocumentVersion, changeReq);
        hlsDocumentVersionService.insertSelective(request, hlsDocumentVersion);
        return hlsDocumentVersion;
    }


    /**
     * 初始化变更map
     *
     * @param transfer
     * @param transferToDivision
     */
    private static void initTransferToDivisionMap(Field transfer, TransferToDivision transferToDivision) {
        Type gType = transfer.getGenericType();
        ParameterizedType pType = (ParameterizedType) gType;
        Type[] actualTypeArguments = pType.getActualTypeArguments();
        if (actualTypeArguments.length > 0) {
            Class clazz = (Class) actualTypeArguments[0];
            String tableName = getTableName(clazz);
            transferToDivisionMap.put(tableName, transferToDivision);
        }

    }

    /**
     * 转换对象,根据传入的对象,返回一个对应的历史对象
     *
     * @param dto
     * @return
     */
    public static BaseDTO getHistoryDto(BaseDTO dto) {
        String tableName = getTableName(dto.getClass());
        Class historyClazz = tableClassMap.get(tableName.concat("_H"));
        BaseDTO historyObj = null;
        try {
            historyObj = (BaseDTO) historyClazz.newInstance();
        } catch (InstantiationException e) {
            e.printStackTrace();
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        }
        //由于现金流表和CshConContract中字段定义不规范,所以单独处理
        if (CON_CONTRACT_CASHFLOW.equalsIgnoreCase(tableName) || dto instanceof CshConContract) {
            List<Field> declaredFields = getAllField(dto.getClass());
            for (Field item : declaredFields) {
                try {
                    String fieldName = item.getName();
                    //String property = BeanUtils.getProperty(dto, fieldName);
                    item.setAccessible(true);
                    Object property = item.get(dto);
                    if (fieldName.contains("_")) {
                        BeanUtils.setProperty(historyObj, underline2Camel(fieldName), property);
                    } else {
                        BeanUtils.setProperty(historyObj, fieldName, property);
                    }

                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                } catch (InvocationTargetException e) {
                    e.printStackTrace();
                }
            }
        } else {
            try {
                BeanUtils.copyProperties(historyObj, dto);
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            }
        }
        cleanStdWhoValue(historyObj);
        return historyObj;
    }

    /**
     * 从历史表dto类型还原回原表dto类型
     *
     * @param fromDto
     * @param toDto
     */
    public static void recoverHistoryDto(BaseDTO fromDto, BaseDTO toDto) {
        Class<? extends BaseDTO> fromDtoClass = fromDto.getClass();
        Class<? extends BaseDTO> toDtoClass = toDto.getClass();
        String historyTableName = getTableName(toDtoClass);
        String originTableName = getTableName(fromDtoClass);
        if (!(historyTableName.endsWith("_H") && historyTableName.contains(originTableName))) {
            return;
        }
        //复制除主键之外的属性
        try {
            String fromPk = BeanUtils.getProperty(fromDto, getIdFliedByAnnotations(fromDtoClass));
            String toPK = BeanUtils.getProperty(toDto, getIdFliedByAnnotations(toDtoClass));
            if (!originTableName.equalsIgnoreCase(CON_CONTRACT_CASHFLOW)) {
                BeanUtils.copyProperties(fromDto, toDto);
            } else {
                List<Field> declaredFields = getAllField(toDto.getClass());
                for (Field item : declaredFields) {
                    try {
                        String fieldName = item.getName();
                        item.setAccessible(true);
                        Object property = item.get(toDto);
                        BeanUtils.setProperty(fromDto, camel2Underline(fieldName), property);
                    } catch (IllegalAccessException e) {
                        e.printStackTrace();
                    } catch (InvocationTargetException e) {
                        e.printStackTrace();
                    }
                }
            }
            BeanUtils.setProperty(fromDto, getIdFliedByAnnotations(fromDtoClass), fromPk);
            cleanStdWhoValue(fromDto);
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        } catch (NoSuchMethodException e) {
            e.printStackTrace();
        }

    }

    /**
     * 清空who字段的值
     *
     * @param dto
     */
    private static void cleanStdWhoValue(BaseDTO dto) {
        dto.setObjectVersionNumber(null);
        dto.setCreatedBy(null);
        dto.setCreationDate(null);
        dto.setLastUpdateDate(null);
        dto.setLastUpdatedBy(null);
        dto.setLastUpdateLogin(null);
    }

    /**
     * 获取类上的table值
     *
     * @param clazz
     * @return
     */
    public static String getTableName(Class clazz) {
        Table[] annotationsByType = (Table[]) clazz.getAnnotationsByType(Table.class);
        Optional<String> tableName = Optional.of(new String());
        if (annotationsByType.length > 0) {
            tableName = Optional.ofNullable(Arrays.asList(annotationsByType).stream().findFirst().get().name());
        }
        return tableName.get().toUpperCase();
    }

    /**
     * 取到类型中的所有的实体类映射字段,带Transient注解的字段除外
     * 如果出现重复字段,则在返回的list中靠前的为子类属性,可根据需求去除重复
     *
     * @param param
     * @return
     */
    public static List<Field> getAllField(Class param) {
        List<Field> list = new ArrayList<>();
        for (Class clazz = param; clazz != null; clazz = clazz.getSuperclass()) {
            Field[] declaredFields = clazz.getDeclaredFields();
            for (Field item : declaredFields) {
                if (item.getAnnotationsByType(Transient.class).length == 0){
                    list.add(item);
                }
            }
        }
        return list;
    }

    /**
     * 取出带id注解的字段
     *
     * @param param
     * @return
     */
    public static String getIdFliedByAnnotations(Class param) {
        List<Field> allFieLd = getAllField(param);
        String idFiledName = null;
        for (Field filed : allFieLd) {
            Id[] annotationsByType = filed.getAnnotationsByType(Id.class);
            if (annotationsByType.length > 0) {
                idFiledName = filed.getName();
                break;
            }
        }
        return idFiledName;
    }

    /**
     * 下划线转驼峰
     *
     * @param param
     * @return
     */
    public static String underline2Camel(String param) {
        if (param == null || "".equals(param.trim())) {
            return "";
        }
        int len = param.length();
        StringBuilder sb = new StringBuilder(len);
        for (int i = 0; i < len; i++) {
            char c = param.charAt(i);
            if (c == UNDERLINE) {
                if (++i < len) {
                    sb.append(Character.toUpperCase(param.charAt(i)));
                }
            } else {
                sb.append(c);
            }
        }
        return sb.toString();
    }


    /**
     * 驼峰法转下划线
     */
    public static String camel2Underline(String parm) {
        if (parm == null || "".equals(parm.trim())) {
            return "";
        }
        parm = String.valueOf(parm.charAt(0)).toUpperCase().concat(parm.substring(1));
        StringBuffer sb = new StringBuffer();
        Pattern pattern = Pattern.compile("[A-Z]([a-z\\d]+)?");
        Matcher matcher = pattern.matcher(parm);
        while (matcher.find()) {
            String word = matcher.group();
            sb.append(word.toLowerCase());
            sb.append(matcher.end() == parm.length() ? "" : "_");
        }
        return sb.toString();
    }


    @Override
    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        //初始化TransferToDivisionMap
        Class clazz = getClass();
        List<Field> fields = new ArrayList<>();
        //取出当前类以及其父类中的所有定义的属性
        for (; clazz != null; clazz = clazz.getSuperclass()) {
            Field[] declaredFields = clazz.getDeclaredFields();
            Arrays.asList(declaredFields).forEach(fields::add);
        }
        fields.forEach(item -> {
            try {
                Object o = item.get(this);
                item.setAccessible(true);
                if (o instanceof TransferToDivision) {//只把类型为TransferToDivision的对象放入到map中
                    initTransferToDivisionMap(item, (TransferToDivision) o);
                }
            } catch (IllegalAccessException e) {
                e.printStackTrace();
                logger.error("please check:{}'s permission", item.getName());
            }
        });
        //初始化相关的Map
        Map<String, IBaseService> beansOfType = applicationContext.getBeansOfType(IBaseService.class);
        beansOfType.forEach((k, v) -> {
            Class serviceClazz = null;
            if (!AopUtils.isAopProxy(v)) {//判断是否为代理对象
                serviceClazz = v.getClass();
            } else {
                serviceClazz = AopUtils.getTargetClass(v);
            }
            Type genType = serviceClazz.getGenericSuperclass();
            if (serviceClazz.isAssignableFrom(CshConContractServiceImpl.class)) {
                return;
            }
            if (genType instanceof ParameterizedType) {
                Type[] params = ((ParameterizedType) genType).getActualTypeArguments();
                if (params.length > 0) {
                    String tableName = getTableName((Class) params[0]);
                    tableServiceMap.put(tableName, v);
                    tableClassMap.put(tableName, (Class) params[0]);
                }
            }
        });
        logger.info("tableClassMap init Complete");
    }

    /**
     * 变更申请对象转换接口
     *
     * @param <T>
     */
    public interface TransferToDivision<T> {
        void transform(HlsCusHlsDocumentVersion documentVersion, T t);
    }

}
