package hls.core.job.service.impl;

import com.hand.hap.core.IRequest;
import com.hand.hap.core.exception.FieldRequiredException;
import com.hand.hap.job.AbstractJob;
import com.hand.hap.job.dto.JobCreateDto;
import com.hand.hap.job.exception.JobException;
import hls.core.job.dto.*;
import hls.core.job.mapper.SchScheduleMapper;
import hls.core.job.mapper.SchTaskMapper;
import hls.core.job.mapper.TaskParamMapper;
import hls.core.job.service.IInformationParaService;
import hls.core.job.service.ISchTaskService;
import org.quartz.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import static com.hand.hap.job.exception.JobException.JOB_EXCEPTION;
import static com.hand.hap.job.exception.JobException.NOT_SUB_CLASS;

/**
 * 由于quartz本身自带事物,所以为了避免事物嵌套,所以应
 * 该把job的创建方法和information表的保存分开
 * Created by fjm on 2017/7/28.
 */
@Service
public class HlsJobCreateService {

    private final Logger logger = LoggerFactory.getLogger(this.getClass());

    @Autowired
    private ISchTaskService service;

    @Autowired
    private IInformationParaService informationParaService;

    @Autowired
    private Scheduler quartzScheduler;

    @Autowired
    private TaskParamMapper taskParamMapper;

    @Autowired
    private SchTaskMapper mapper;

    @Autowired
    private SchScheduleMapper schScheduleMapper;

    private List<TaskParam> taskParams;//task的参数集合

    private JobTask schTask;//job对应的task对象



    /**
     * 创建一个新的job,但是Job的参数从task_param表中获取
     * @param hlsJobDto
     * @throws ClassNotFoundException
     * @throws SchedulerException
     * @throws JobException
     * @return jobId
     */

    public Long createHlsJob(HlsJobDto hlsJobDto, IRequest requestCtx) throws ClassNotFoundException, SchedulerException, JobException {
        Long taskId = hlsJobDto.getTaskId();
        String schName = hlsJobDto.getScheduleName();
        JobCreateDto jobCreateDto = hlsJobDto.getJobCreateDto();
        JobTask jobTask = null;
        HashMap businessParam = hlsJobDto.getBusinessParam();
        HashMap defineTaskParam = hlsJobDto.getTaskParam();

        HashMap<String,Long> mediumMap = new HashMap();

        if (StringUtils.isEmpty(jobCreateDto.getJobClassName())) {
            throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "jobClassName" }));
        } else if (StringUtils.isEmpty(jobCreateDto.getJobName())) {
            throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "jobName" }));
        } else if (StringUtils.isEmpty(jobCreateDto.getJobGroup())) {
            throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "jobGroup" }));
        } else if (StringUtils.isEmpty(jobCreateDto.getTriggerName())) {
            throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "triggerName" }));
        } else if (StringUtils.isEmpty(jobCreateDto.getTriggerGroup())) {
            throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "triggerGroup" }));
        } else if (StringUtils.isEmpty(jobCreateDto.getTriggerType())) {
            throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "triggerType" }));
        }

        String jobClassName;

        if (taskId == null){
            jobClassName = jobCreateDto.getJobClassName();
        }else {
            jobTask = new JobTask();
            jobTask.setTaskId(taskId);
            jobTask = mapper.selectOne(jobTask);
            jobClassName = jobTask.getJobClassName();
            jobCreateDto.setJobClassName(jobClassName);
        }


        boolean assignableFrom = false;
        Class forName = null;
        try {
            forName = Class.forName(jobClassName);
            assignableFrom = AbstractJob.class.isAssignableFrom(forName);
        } catch (ClassNotFoundException e) {
            if (logger.isErrorEnabled()) {
                logger.error(e.getMessage(), e);
            }
        }
        if (!assignableFrom || forName == null) {
            String name = AbstractJob.class.getName();
            throw new JobException(JOB_EXCEPTION, NOT_SUB_CLASS, new Object[] { jobClassName, name });
        }

        JobBuilder jb = JobBuilder.newJob(forName).withIdentity(jobCreateDto.getJobName(), jobCreateDto.getJobGroup())
                .withDescription(jobCreateDto.getDescription());


        //如果task中定义了参数,把task中的参数放入到JobDataMap中
        if (hasJobData(jobTask,taskId)) {
            JobDataMap data = new JobDataMap();
            taskParams.forEach(item->{
                data.put(item.getParamName(),item.getParamValue());
                mediumMap.put(item.getParamName(),item.getParamId());//用于存放参数名和参数ID
            });
            if (defineTaskParam!=null && defineTaskParam.size()>0){//如果在工作台创建job时,修改了参数的值,那么应该用这个参数来覆盖原参数
                defineTaskParam.forEach((k,v)->data.put((String) k,v));
            }
            jb = jb.usingJobData(data);
        }
        JobDetail jobDetail = jb.build();

        //如果含有业务类型的参数,把业务类型的参数也放入到JobDataMap之中
        if (businessParam!=null){
            JobDataMap map = jobDetail.getJobDataMap();
            businessParam.forEach((k,v)->{
                map.put((String) k,v);
            });
        }

        //把IRequest放入到JobDataMap之中
        if (requestCtx!=null){
            jobDetail.getJobDataMap().put("requestContext",requestCtx);
        }

        TriggerBuilder<Trigger> triggerBuilder = TriggerBuilder.newTrigger()
                .withIdentity(jobCreateDto.getTriggerName(), jobCreateDto.getTriggerGroup()).forJob(jobDetail);

        if (jobCreateDto.getStartTime() != null && jobCreateDto.getStartTime() > 0) {
            triggerBuilder.startAt(new Date(jobCreateDto.getStartTime()));
        }
        if (jobCreateDto.getEndTime() != null && jobCreateDto.getEndTime() > 0) {
            triggerBuilder.endAt(new Date(jobCreateDto.getEndTime()));
        }

        //如果额外定义了开始时间和结束时间,则覆盖定义的scheduler上面的值
        if (hlsJobDto.getDefineStartTime() != null && hlsJobDto.getDefineStartTime()  > 0) {
            triggerBuilder.startAt(new Date(hlsJobDto.getDefineStartTime() ));
        }
        if (hlsJobDto.getDefineEndTime() != null && hlsJobDto.getDefineEndTime()  > 0) {
            triggerBuilder.endAt(new Date(hlsJobDto.getDefineEndTime() ));
        }

        ScheduleBuilder sche = null;

        if ("CRON".equalsIgnoreCase(jobCreateDto.getTriggerType())) {
            if (org.apache.commons.lang.StringUtils.isEmpty(jobCreateDto.getCronExpression())) {
                throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "cronExpression" }));
            }
            if (schName == null) {
                sche = CronScheduleBuilder.cronSchedule(jobCreateDto.getCronExpression());
            }else {

                JobSchedule jobSchedule = new JobSchedule();
                jobSchedule.setScheduleName(schName);
                jobSchedule =  schScheduleMapper.selectOne(jobSchedule);
                String cronExpression = jobSchedule.getCronExpression();
                jobCreateDto.setCronExpression(cronExpression);
                jobCreateDto.setSchedName(schName);
                sche = CronScheduleBuilder.cronSchedule(cronExpression);
            }

        } else if ("SIMPLE".equalsIgnoreCase(jobCreateDto.getTriggerType())) {
            if (jobCreateDto.getStartTime()==null) {
                throw new RuntimeException(new FieldRequiredException("hap.validation.notempty", new Object[] { "startTime could not be null" }));
            }
            sche =  SimpleScheduleBuilder.repeatSecondlyForTotalCount(1);

        } else if ("NOW".equalsIgnoreCase(jobCreateDto.getTriggerType())){

            sche = SimpleScheduleBuilder.simpleSchedule().withIntervalInSeconds(3).withRepeatCount(0);//立即执行一个简单job,只执行一次
            //sche = SimpleScheduleBuilder.repeatSecondlyForTotalCount(1);

        }
        Trigger trigger = triggerBuilder.withSchedule(sche).build();
        quartzScheduler.scheduleJob(jobDetail, trigger);//将job和trigger放入调度器之中(如果调度器是启动状态,那么job会开始执行)

        //将job的创建信息,存入JobInformation中
        Long jobId = service.saveJobInformation(hlsJobDto,requestCtx);
        //
        if (defineTaskParam!=null && defineTaskParam.size()>0){//如果在工作台创建job时,修改了参数的值,那么应该用这个参数来覆盖原参数
            defineTaskParam.forEach((k,v)->{
                InformationPara para = new InformationPara();
                para.setJobId(jobId);
                para.setParameterDesc((String) k);
                para.setParameterValue((String) v);
                para.setParameterName((String) k);
                para.setTaskParameterId(mediumMap.get(k));
                informationParaService.self().insertSelective(requestCtx,para);
            });
        }

        return jobId;//返回job的id
    }

    /**
     * 判断是否有jobData
     * @param task
     * @return
     */
    private boolean hasJobData(JobTask task, Long taskId) {
        if (task!=null){
            TaskParam taskParam = new TaskParam();
            taskParam.setTaskId(task.getTaskId());
            taskParams = taskParamMapper.select(taskParam);
        }else {
            //先获取schTask对象
            JobTask obj = new JobTask();
            obj.setTaskId(taskId);
            this.schTask = mapper.selectOne(obj);

            //获取一个task下对应的taskParam参数集
            TaskParam taskParam = new TaskParam();
            taskParam.setTaskId(schTask.getTaskId());
            taskParams = taskParamMapper.select(taskParam);
        }
        return taskParams != null && !taskParams.isEmpty();
    }


}
