package hls.core.svc.components;

import hls.core.svc.beans.*;
import hls.core.svc.utils.ParamUtils;
import jdk.nashorn.api.scripting.NashornScriptEngine;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Assert;
import org.springframework.web.context.support.XmlWebApplicationContext;

import javax.script.ScriptEngineManager;
import javax.script.ScriptException;
import javax.servlet.http.HttpServletRequest;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-12-27
 * Time: 11:47
 */
@Component
public class ServiceExecutor {
    private NashornScriptEngine engine;
    private String webPath;
    private ModelExecutor modelExecutor;
    private Logger logger = LoggerFactory.getLogger(getClass());

    @Autowired
    public ServiceExecutor(XmlWebApplicationContext context, ModelExecutor modelExecutor) {
        engine = (NashornScriptEngine) new ScriptEngineManager().getEngineByName("nashorn");
        this.modelExecutor = modelExecutor;
        try {
            URL resource = context.getServletContext().getResource("/WEB-INF");
            Assert.notNull(resource, "can not found /WEB-INF");
            webPath = resource.getPath() + "/js";
        } catch (MalformedURLException e) {
            if (logger.isDebugEnabled()) {
                logger.debug("ServiceExecutor init failed", e);
            }
        }

    }

    @Transactional
    public void execute(Service service) {
        Assert.notNull(service, "service can not be null");
        ServerScript serverScript = service.getServerScript();
        InitProcedure initProcedure = service.getInitProcedure();
        if (serverScript != null) {
            executeScript(service);
        } else if (initProcedure != null) {
            // execute othrer tags if <script> not exists
            executeProcedure(service);
        }
    }

    private void executeProcedure(Service service) {
        Assert.notNull(service, "service can not be null");
        InitProcedure initProcedure = service.getInitProcedure();
        Assert.notNull(initProcedure, "initProcedure can not be null");

        // Create
        ModelInsert modelInsert = initProcedure.getModelInsert();
        if (modelInsert != null) {
            modelInsert.execute(modelExecutor);
        }

        // Retrieve
        ModelQuery modelQuery = initProcedure.getModelQuery();
        if (modelQuery != null) {
            modelQuery.execute(modelExecutor);
        }

        // Update
        ModelUpdate modelUpdate = initProcedure.getModelUpdate();
        if (modelUpdate != null) {
            modelUpdate.execute(modelExecutor);
        }

        // Delete
        ModelDelete modelDelete = initProcedure.getModelDelete();
        if (modelDelete != null) {
            modelDelete.execute(modelExecutor);
        }
    }

    /**
     * execute js part in service file
     *
     * @param service parsed from service file
     */
    private void executeScript(Service service) {
        if (service == null) {
            return;
        }
        ServerScript serverScript = service.getServerScript();
        if (serverScript == null) {
            return;
        }
        String reference = serverScript.getReference();

        prepareEngine();
        // load import js
        if (StringUtils.isNotBlank(reference)) {
            String[] references = StringUtils.split(reference, ",");
            for (String ref : references) {
                try {
                    engine.eval(readJsPath(ref));
                } catch (Exception e) {
                    if (logger.isDebugEnabled()) {
                        logger.debug("import js error", e);
                    }
                    if (logger.isTraceEnabled()) {
                        logger.trace(ref);
                    }
                }
            }
        }

        // execute script content
        if (StringUtils.isNotBlank(serverScript.getContent())) {
            try {
                engine.eval(serverScript.getContent());
            } catch (ScriptException e) {
                if (logger.isDebugEnabled()) {
                    logger.debug("execute service error", e);
                }
                if (logger.isTraceEnabled()) {
                    logger.trace(serverScript.getContent());
                }
            }
        }
    }

    /**
     * get target js path
     *
     * @param js js file name, locate at /WEB-INF/js
     * @return
     */
    public Reader readJsPath(String js) throws FileNotFoundException {
        Assert.hasLength(js, "js path can not be empty");
        String path = webPath + StringUtils.replace(js, ".", File.separator);
        File file = new File(path);
        if (!file.exists()) {
            throw new IllegalArgumentException(js + " js not exist");
        } else {
            return new FileReader(file);
        }
    }

    /**
     * prepare engine global variable like $BM,$DB
     */
    private void prepareEngine() {
        try {
            HttpServletRequest currentHttpRequest = ParamUtils.getCurrentHttpRequest();
            engine.put("$REQUEST", currentHttpRequest);
            engine.put("$DB", modelExecutor);
            engine.eval("var $BM = Java.type('com.hand.hls.web.svc.utils.ModelParser');");
        } catch (ScriptException e) {
            logger.error("prepare engine error", e);
        }
    }
}
