package hls.core.svc.components;

import com.alibaba.fastjson.JSON;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.RowCallbackHandler;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.jdbc.support.KeyHolder;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import java.sql.ResultSetMetaData;
import java.util.*;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-12-20
 * Time: 14:40
 */
@Component
public class SqlExecutor {
    private final NamedParameterJdbcTemplate jdbcTemplate;
    private Logger logger = LoggerFactory.getLogger(getClass());

    @Autowired
    public SqlExecutor(NamedParameterJdbcTemplate jdbcTemplate) {
        this.jdbcTemplate = jdbcTemplate;
    }

    public List<Map<String, Object>> executeQuery(String sql, Map<String, Object> params) {
        List<Map<String, Object>> results = new ArrayList<>();
        executeQuery(sql, params, resultSet -> {
            HashMap<String, Object> map = new HashMap<>();
            if (resultSet == null) {
                return;
            }
            ResultSetMetaData metaData = resultSet.getMetaData();
            int columnCount = metaData.getColumnCount();
            for (int i = 1; i <= columnCount; i++) {
                map.put(metaData.getColumnLabel(i), resultSet.getObject(i));
            }
            results.add(map);
        });
        logger.debug("query result: {}", JSON.toJSONString(results));
        return results;
    }

    public void executeQuery(String sql, Map<String, Object> params, RowCallbackHandler rowCallbackHandler) {
        if (logger.isDebugEnabled()) {
            logger.debug("\nsql: {}\nparams: {}", sql, JSON.toJSONString(params));
        }
        Assert.notNull(sql, "sql can not be null");
        Assert.notNull(rowCallbackHandler, "rowCallbackHandler can not be null");
        jdbcTemplate.query(sql, params, rowCallbackHandler);
    }

    /**
     * execute update statement
     *
     * @param sql    update sql
     * @param params update params
     * @return updated keys when updated data, or null if update nothing
     */
    public Map<String, Object> executeUpdate(String sql, Map<String, Object> params) {
        if (logger.isDebugEnabled()) {
            logger.debug("\nsql: {}\nparams: {}", sql, JSON.toJSONString(params));
        }
        return executeUpdate(sql, new MapSqlParameterSource(params));
    }

    /**
     * execute update statement
     *
     * @param sql             update sql
     * @param parameterSource update params
     * @return updated keys when updated data, or null if update nothing
     */
    public Map<String, Object> executeUpdate(String sql, SqlParameterSource parameterSource) {
        Assert.notNull(sql, "sql can not be null");
        Assert.notNull(parameterSource, "update can not be null");

        KeyHolder keyHolder = new GeneratedKeyHolder();
        int update = jdbcTemplate.update(sql, parameterSource, keyHolder);
        if (logger.isWarnEnabled()) {
            logger.warn("\nupdate sql:{}\nupdated:{}", sql, update);
        }
        if (update > 0) {
            // updated data
            return keyHolder.getKeys();
        }
        return Collections.emptyMap();
    }

    public Map<String, Object> executeInsert(String sql, Map<String, Object> params) {
        return executeUpdate(sql, params);
    }

    public int executeDelete(String sql, Map<String, Object> params) {
        return jdbcTemplate.update(sql, params);
    }

}
