package hls.core.svc.utils;

import com.hand.hap.core.web.view.XMap;
import com.hand.hap.core.web.view.XMapParser;
import com.hand.hap.core.web.view.XMapUtil;
import hls.core.svc.beans.BusinessModel;
import hls.core.svc.beans.Field;
import hls.core.svc.beans.PkField;
import hls.core.svc.beans.QueryField;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.Assert;
import org.springframework.util.CollectionUtils;
import org.xmlpull.v1.XmlPullParserException;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-12-19
 * Time: 15:49
 */
public class ModelParser {

    private static Logger logger = LoggerFactory.getLogger(ModelParser.class);

    /**
     * duplicate for js file
     *
     * @param path bm location,separate with dot(.)
     * @return
     * @throws IOException
     * @throws XmlPullParserException
     */
    public static BusinessModel load(String path) throws IOException, XmlPullParserException {
        return parseModel(path);
    }
    /**
     * parse bm file to BusinessModel
     *
     * @param bmPath bm location,separate with dot(.)
     * @return parsed BusinessModel object
     * @throws IOException
     * @throws XmlPullParserException
     */
    public static BusinessModel parseModel(String bmPath) throws IOException, XmlPullParserException {
        if (StringUtils.isEmpty(bmPath)) {
            throw new IllegalArgumentException("bm path cannot be empty");
        }
        // get bm resource
        String path = StringUtils.replace(bmPath, ".", File.separator);
        URL resource = ModelParser.class.getResource("/bm/" + path + ".bm");
        if (resource == null) {
            throw new IllegalArgumentException("cannot found bm[" + bmPath + "]");
        }
        // read bm as stream
        InputStream inputStream = resource.openStream();
        // parse to XMap
        XMap xMap = XMapParser.parseStream(inputStream);
        return ModelParser.parseModel(xMap);
    }

    /**
     * parse BusinessModel from xMap
     *
     * @param xMap xMap contains properties and sub-nodes
     * @return parsed BusinessModel object
     */
    public static BusinessModel parseModel(XMap xMap) {
        Assert.notEmpty(xMap, "bm has no content");
        if (!"model".equalsIgnoreCase(xMap.getName())) {
            throw new IllegalArgumentException("not a bm");
        }
        BusinessModel bm = new BusinessModel();
        HashMap<String, Object> map = new HashMap<>();
        XMapUtil.copyAttributes(xMap, map);
        copyXmap2Model(xMap, bm);
        return bm;
    }

    private static void copyXmap2Model(XMap xMap, BusinessModel bm) {
        Assert.notNull(xMap, "illegal parameter xMap");
        Assert.notNull(bm, "illegal parameter bm");
        if (!"model".equalsIgnoreCase(xMap.getName())) {
            throw new IllegalArgumentException("not a bm");
        }
        // properties
        bm.setBaseTable(xMap.getString("baseTable"));
        bm.setAlias(xMap.getString("alias"));
        bm.setDataSourceName(xMap.getString("dataSourceName"));
        bm.setExtendMode(xMap.getString("extendMode"));
        bm.setDefaultOrderBy(xMap.getString("defaultOrderBy"));
        bm.setAccessControlMode(xMap.getString("accessControlMode"));
        bm.setMaxPageSize(xMap.getString("maxPageSize"));
        bm.setDefaultDisplayField(xMap.getString("defaultDisplayField"));
        //FIXME: extend field
        String extend = xMap.getString("extend");
        String needAccessControl = xMap.getString("needAccessControl");
        String allowFetchAll = xMap.getString("allowFetchAll");

        bm.setNeedAccessControl(!"false".equalsIgnoreCase(needAccessControl));
        bm.setAllowFetchAll(!"false".equalsIgnoreCase(allowFetchAll));

        // sub nodes
        XMap fields = xMap.getChild("fields");
        XMap queryFields = xMap.getChild("queryFields");
        XMap pkFields = xMap.getChild("pkFields");

        Map<String, Field> fieldMap = copyFields2Model(fields, bm);

        if(CollectionUtils.isEmpty(fieldMap)){
            if(logger.isWarnEnabled()){
                logger.warn("BusinessModel(name={}) has no fields, won't config queryFields and PkFields.", bm.getBaseTable());
            }
        }else {
            copyQueryFields2Model(queryFields, bm, fieldMap);
            copyXmap2PkFields(pkFields, bm, fieldMap);
        }
    }

    private static void copyXmap2PkFields(XMap pkFields, BusinessModel bm, Map<String, Field> fieldMap) {
        Assert.notNull(pkFields, "illegal parameter pkFields");
        Assert.notNull(bm, "illegal parameter bm");
        if (!"pkFields".equalsIgnoreCase(pkFields.getName())) {
            throw new IllegalArgumentException("not a PkFields");
        }
        // get children (must be pkField)
        List<XMap> fields = pkFields.getChildren();
        if (fields != null) {
            fields.stream()
                    .filter(field -> field != null && "pkField".equalsIgnoreCase(field.getName()))
                    .forEach(field -> {
                        PkField pkField = new PkField();
                        copyXmap2PkField(field, pkField, fieldMap);
                        List<PkField> list = bm.getPkFields();
                        if (list == null) {
                            list = new ArrayList<>();
                            bm.setPkFields(list);
                        }
                        list.add(pkField);
                    });
        }
    }

    private static void copyXmap2PkField(XMap xMap, PkField pkField, Map<String, Field> fields) {
        Assert.notNull(xMap, "illegal parameter xMap");
        Assert.notNull(pkField, "illegal parameter pkField");
        if (!"pkField".equalsIgnoreCase(xMap.getName())) {
            throw new IllegalArgumentException("not a pkField");
        }
        // only 1 property
        String name = xMap.getString("name");
        pkField.setName(name);
        if (fields != null) {
            if (fields.containsKey(name)) {
                pkField.setField(fields.get(name));
            }
        }
    }

    private static void copyQueryFields2Model(XMap queryFields, BusinessModel bm, Map<String, Field> fieldMap) {
        Assert.notNull(queryFields, "illegal parameter queryFields");
        Assert.notNull(bm, "illegal parameter bm");
        if (!"queryFields".equalsIgnoreCase(queryFields.getName())) {
            throw new IllegalArgumentException("not a QueryFields");
        }
        List<XMap> fields = queryFields.getChildren();
        if (fields != null) {
            fields.stream()
                    .filter(field -> field != null && "queryField".equalsIgnoreCase(field.getName()))
                    .forEach(field -> {
                        QueryField queryField = new QueryField();
                        copyXmap2QueryField(field, queryField, fieldMap);
                        List<QueryField> list = bm.getQueryFields();
                        if (list == null) {
                            list = new ArrayList<>();
                            bm.setQueryFields(list);
                        }
                        list.add(queryField);
                    });
        }
    }

    private static void copyXmap2QueryField(XMap xMap, QueryField queryField, Map<String, Field> fieldMap) {
        Assert.notNull(xMap, "illegal parameter xMap");
        Assert.notNull(queryField, "illegal parameter queryField");
        if (!"queryField".equalsIgnoreCase(xMap.getName())) {
            throw new IllegalArgumentException("not a queryField");
        }
        queryField.setName(xMap.getString("name"));
        queryField.setQueryOperator(xMap.getString("queryOperator"));
        queryField.setQueryExpression(xMap.getString("queryExpression"));
        queryField.setDataType(xMap.getString("dataType"));
        queryField.setRequired(!"false".equalsIgnoreCase(xMap.getString("required")));
        String field = xMap.getString("field");
        if (fieldMap != null) {
            if (fieldMap.containsKey(field)) {
                queryField.setField(fieldMap.get(field));
            }
        }
    }

    private static Map<String, Field> copyFields2Model(XMap xMap, BusinessModel bm) {
        Assert.notNull(xMap, "illegal parameter fields");
        Assert.notNull(bm, "illegal parameter bm");
        if (!"fields".equalsIgnoreCase(xMap.getName())) {
            throw new IllegalArgumentException("not a Fields");
        }
        List<XMap> fields = xMap.getChildren();
        if (fields != null) {
            Map<String, Field> fieldMap = new HashMap<>(fields.size());
            fields.stream()
                    .filter(field -> field != null && "field".equalsIgnoreCase(field.getName()))
                    .forEach(field -> {
                        Field field1 = new Field();
                        copyXmap2Field(field, field1);
                        List<Field> list = bm.getFields();
                        if (list == null) {
                            list = new ArrayList<>();
                            bm.setFields(list);
                        }
                        fieldMap.put(field1.getName(), field1);
                        list.add(field1);
                    });
            return fieldMap;
        }
        return null;
    }

    private static void copyXmap2Field(XMap xMap, Field field) {
        Assert.notNull(xMap, "illegal parameter xMap");
        Assert.notNull(field, "illegal parameter field");
        if (!"field".equalsIgnoreCase(xMap.getName())) {
            throw new IllegalArgumentException("not a field");
        }
        field.setName(xMap.getString("name"));
        field.setDataType(xMap.getString("dataType"));
        field.setPrompt(xMap.getString("prompt"));
        field.setPhysicalName(xMap.getString("physicalName"));
        field.setDatabaseType(xMap.getString("databaseType"));
        field.setRequired(!"false".equalsIgnoreCase(xMap.getString("required")));
        field.setDefaultValue(xMap.getString("defaultValue"));
        field.setParameterPath(xMap.getString("parameterPath"));
        field.setExpression(xMap.getString("expression"));
        field.setForQuery(!"false".equalsIgnoreCase(xMap.getString("forQuery")));
        field.setForSelect(!"false".equalsIgnoreCase(xMap.getString("forSelect")));
        field.setForInsert(!"false".equalsIgnoreCase(xMap.getString("forInsert")));
        field.setForUpdate(!"false".equalsIgnoreCase(xMap.getString("forUpdate")));
        field.setInsertExpression(xMap.getString("insertExpression"));
        field.setUpdateExpression(xMap.getString("updateExpression"));
        field.setQueryExpression(xMap.getString("queryExpression"));
        field.setMultiLanguage(xMap.getString("multiLanguage"));
        field.setMultiLanguageDescField(xMap.getString("multiLanguageDescField"));
        field.setForceUpdate(xMap.getString("forceUpdate"));
        field.setLookupCode(xMap.getString("lookupCode"));
        field.setLookupField(xMap.getString("lookupField"));
    }
}
