package hls.core.svc.utils;

import com.hand.hap.core.web.view.XMap;
import com.hand.hap.core.web.view.XMapParser;
import hls.core.svc.beans.*;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.Assert;

import java.io.File;
import java.io.InputStream;
import java.net.URL;
import java.util.List;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-12-25
 * Time: 9:02
 */
public class ServiceParser {
    private static Logger logger = LoggerFactory.getLogger(ServiceParser.class);
    private static String SERVICE_FILE_SUFFIX = ".service";

    public static Service parseService(String servicePath) throws Exception {
        if (StringUtils.isEmpty(servicePath)) {
            throw new IllegalArgumentException("servicePath path cannot be empty");
        }
        // get service resource
        String path = StringUtils.replace(servicePath, ".", File.separator);
        URL resource = ModelParser.class.getResource("/service/" + path + SERVICE_FILE_SUFFIX);
        if (resource == null) {
            throw new IllegalArgumentException("cannot found service[" + servicePath + "]");
        }
        // read service as stream
        InputStream inputStream = resource.openStream();
        // parse to XMap
        return parseService(XMapParser.parseStream(inputStream));
    }

    public static Service parseService(XMap xMap) {
        Assert.notNull(xMap, "xMap can not be null");
        if(!"service".equals(xMap.getName())){
            throw new IllegalArgumentException("xMap is not a service");
        }
        Service service = new Service();

        service.setTag(xMap.getString("tag"));
        service.setCacheEnabled(xMap.getBoolean("cacheEnabled"));
        service.setCacheKey(xMap.getString("cacheKey"));
        service.setCheckSessionLock(xMap.getBoolean("checkSessionLock"));
        service.setLockErrorMessage(xMap.getString("lockErrorMessage"));
        service.setLockKey(xMap.getString("lockKey"));
        service.setTrace(xMap.getBoolean("trace"));

        List<XMap> children = xMap.getChildren();
        if (children != null) {
            for (XMap child : children) {
                if ("script".equals(child.getName())) {
                    ServerScript serverScript = new ServerScript();
                    serverScript.setContent(child.getText());
                    serverScript.setReference(child.getString("import"));
                    service.setServerScript(serverScript);
                } else if ("serviceOutput".equals(child.getName())) {
                    service.setServiceOutput(child.getString("output"));
                } else if ("initProcedure".equals(child.getName())) {
                    InitProcedure initProcedure = new InitProcedure();
                    copyXmap2InitProcedure(child, initProcedure);
                    service.setInitProcedure(initProcedure);
                }
            }
        }
        return service;
    }

    private static void copyXmap2InitProcedure(XMap xMap, InitProcedure initProcedure) {
        if (xMap == null || initProcedure == null || !"initProcedure".equals(xMap.getName())) {
            throw new IllegalArgumentException("illegal arguments");
        }
        List<XMap> children = xMap.getChildren();
        if (children == null) {
            return;
        }
        for (XMap child : children) {
            switch (child.getName()) {
                case "modelUpdate":
                    ModelUpdate modelUpdate = new ModelUpdate();
                    modelUpdate.setModel(child.getString("model"));
                    modelUpdate.setUpdatePassedFieldOnly(child.getBoolean("updatePassedFieldOnly"));
                    modelUpdate.setSourcePath(child.getString("sourcePath"));
                    initProcedure.setModelUpdate(modelUpdate);
                    break;
                case "modelInsert":
                    ModelInsert modelInsert = new ModelInsert();
                    modelInsert.setModel(child.getString("model"));
                    modelInsert.setSourcePath(child.getString("sourcePath"));
                    initProcedure.setModelInsert(modelInsert);
                    break;
                case "modelDelete":
                    ModelDelete modelDelete = new ModelDelete();
                    modelDelete.setSourcePath(child.getString("sourcePath"));
                    modelDelete.setModel(child.getString("model"));
                    initProcedure.setModelDelete(modelDelete);
                    break;
                case "modelQuery":
                    ModelQuery modelQuery = new ModelQuery();
                    modelQuery.setSourcePath(child.getString("sourcePath"));
                    modelQuery.setModel(child.getString("model"));
                    modelQuery.setAutoCount(child.getBoolean("autoCount"));
                    modelQuery.setDefaultWhereClause(child.getString("defaultWhereClause"));
                    modelQuery.setFetchAll(child.getBoolean("fetchAll"));
                    modelQuery.setQueryOrderBy(child.getString("queryOrderBy"));
                    modelQuery.setRootPath(child.getString("rootPath"));
                    modelQuery.setRecordName(child.getString("recordName"));
                    initProcedure.setModelQuery(modelQuery);
                    break;
                default:
                    logger.warn("{} : not match any tag", child.getName());
            }
        }

    }

}
