package hls.core.svc.utils;

import hls.core.svc.beans.BusinessModel;
import hls.core.svc.beans.Field;
import hls.core.svc.beans.PkField;
import hls.core.svc.beans.QueryField;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.util.Assert;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * Created with IntelliJ IDEA.
 * User: yang
 * Date: 2017-12-19
 * Time: 19:50
 */
public class SqlCreator {
    private BusinessModel businessModel;
    private String tableName;

    public SqlCreator(BusinessModel model) {
        this.businessModel = model;
        tableName = businessModel.getBaseTable();
        if (StringUtils.isBlank(tableName)) {
            throw new IllegalArgumentException("baseTable can not be blank");
        }
    }

    /**
     *  Query Part
     */


    /**
     * query all fields without params
     *
     * @return
     */
    public String getSelect() {
        String select = createSelect();
        String from = createFrom();
        String where = createWhere();
        return select + from + where;
    }

    /**
     * query all fields without params
     *
     * @return
     */
    public String getSelectAll() {
        return getSelectAll(false);
    }

    /**
     * query all fields or specified fields without params
     *
     * @param useFields use specified fields as select part  or not
     * @return
     */
    public String getSelectAll(boolean useFields) {
        String select = "select * ";
        if (useFields) {
            select = createSelect();
        }
        String from = createFrom();
        return select + from;
    }

    /**
     * query all fields with params
     *
     * @param params params to query
     * @return
     */
    public String getSelectAllWithParam(Map<String, Object> params) {
        return getSelectAllWithParam(params, false);
    }

    /**
     * query all fields or specified fields with params
     *
     * @param params    params to query
     * @param useFields use specified fields as select part or not
     * @return
     */
    public String getSelectAllWithParam(Map<String, Object> params, boolean useFields) {
        String select = "select * ";
        if (useFields) {
            select = createSelect();
        }
        String from = createFrom();
        String where = createWhere(businessModel.getFields(), params);
        return select + from + where;
    }

    public String getSelectOperation() {
        //todo: operation query
        return null;
    }

    /// Update part

    /**
     * update field specified in fields marked forUpdate
     *
     * @return update sql
     */
    public String getUpdateByPrimaryKey() {
        return getUpdateByPrimaryKeyWithParams(null);
    }


    /**
     * update field specified in fields marked forUpdate or specified in params
     *
     * @param params
     * @return
     */
    public String getUpdateByPrimaryKeyWithParams(Map<String, Object> params) {
        List<PkField> pkFields = businessModel.getPkFields();
        List<Field> fields = businessModel.getFields();
        if (CollectionUtils.isEmpty(pkFields)) {
            throw new IllegalArgumentException("must have at least 1 pkField");
        }
        if (CollectionUtils.isEmpty(fields)) {
            throw new IllegalArgumentException("must have at least 1 fields");
        }
        List<String> whereList = pkFields.stream().map(pkField -> {
            Field field = pkField.getField();
            String name = field.getName();
            String physicalName = field.getPhysicalName();
            if (StringUtils.isBlank(physicalName)) {
                throw new IllegalArgumentException("pkField ref field must have 'physicalName' attribute");
            }
            return physicalName + " = :" + name;
        }).collect(Collectors.toList());
        if (CollectionUtils.isEmpty(whereList)) {
            throw new IllegalArgumentException("must have at least 1 valid pkField");
        }
        ArrayList<String> fieldList = new ArrayList<>(fields.size());
        ArrayList<String> valueList = new ArrayList<>(fields.size());
        fields.stream()
                .filter(field -> {
                    if (params != null) {
                        return params.containsKey(field.getName());
                    } else {
                        return field.getForUpdate() == null || field.getForUpdate();
                    }
                })
                .forEach(field -> {
                    String updateExpression = field.getUpdateExpression();
                    String physicalName = field.getPhysicalName();
                    String name = field.getName();
                    if (StringUtils.isNotBlank(updateExpression) && (params == null || !params.containsKey(name))) {
                        valueList.add(updateExpression);
                    } else {
                        valueList.add(":" + name);
                    }
                    fieldList.add(physicalName);
                });
        if (CollectionUtils.isEmpty(fieldList) || CollectionUtils.isEmpty(valueList) || fieldList.size() != valueList.size()) {
            throw new IllegalArgumentException("fields for update not match");
        }
        StringBuilder updateBuilder = new StringBuilder("update ").append(tableName).append(" set ");
        for (int i = 0; i < fieldList.size(); i++) {
            if (i > 0) {
                updateBuilder.append(",");
            }
            updateBuilder.append(fieldList.get(i)).append(" = ").append(valueList.get(i));
        }
        updateBuilder.append(" where ").append(StringUtils.join(whereList, " and "));
        return updateBuilder.toString();
    }


    // delete part


    /**
     * parse delete statement by using all pkField
     *
     * @return delete sql
     */
    public String getDeleteByPrimaryKey() {
        List<PkField> pkFields = businessModel.getPkFields();
        if (CollectionUtils.isEmpty(pkFields)) {
            throw new IllegalArgumentException("must have at least 1 pkField");
        }
        List<String> whereList = pkFields.stream().map(pkField -> {
            Field field = pkField.getField();
            String physicalName = field.getPhysicalName();
            if (StringUtils.isBlank(physicalName)) {
                throw new IllegalArgumentException("pkField ref field must have 'physicalName' attribute");
            }
            return physicalName + " = :" + field.getName();
        }).collect(Collectors.toList());
        if (CollectionUtils.isEmpty(whereList)) {
            throw new IllegalArgumentException("must have at least 1 valid pkField");
        }
        return "delete from " + tableName + " where " + StringUtils.join(whereList, " and ");
    }

    // insert part

    /**
     * parse insert sql with all fields specified with forInsert
     *
     * @return insert all fields sql
     */
    public String getInsert() {
        return getInsert(null);
    }

    /**
     * parse insert sql with all fields specified with forInsert or fields in params' key
     *
     * @return insert all fields sql if params is {@link null} or fields in params' keys
     */
    public String getInsert(Map<String, Object> params) {
        List<Field> fields = businessModel.getFields();
        Assert.notEmpty(fields, "fields can not be empty");
        ArrayList<String> fieldList = new ArrayList<>(fields.size());
        ArrayList<String> valueList = new ArrayList<>(fields.size());

        for (Field field : fields) {
            Boolean forInsert = field.getForInsert();
            String name = field.getName();
            if (params != null && !params.containsKey(name)) {
                continue;
            }
            if (forInsert != null && !forInsert) {
                continue;
            }
            String insertExpression = field.getInsertExpression();
            String physicalName = field.getPhysicalName();
            if (StringUtils.isNotBlank(insertExpression)) {
                valueList.add(insertExpression);
            } else {
                valueList.add(":" + name);
            }
            fieldList.add(physicalName);

        }
        return "insert into " + tableName + "(" + StringUtils.join(fieldList, ",") + ") values(" + StringUtils.join(valueList, ",") + ")";
    }


    private String createWhere(List<Field> fields, Map<String, Object> params) {
        StringBuilder whereBuilder = new StringBuilder(" where 1=1 ");
        if (CollectionUtils.isEmpty(fields) || params == null || params.size() < 1) {
            return whereBuilder.toString();
        }
        String alias = businessModel.getAlias();
        if (StringUtils.isBlank(alias)) {
            alias = tableName;
        }
        final String finalAlias = alias;
        fields.stream().filter(field -> params.containsKey(field.getName())).forEach(field -> {
            whereBuilder.append(" and ").append(finalAlias).append(".").append(field.getPhysicalName()).append("=:").append(field.getName());
        });
        return whereBuilder.toString();
    }


    /**
     * parse query where part by using businessModel
     *
     * @return where parse String
     */
    private String createWhere() {
        StringBuilder whereBuilder = new StringBuilder(" where 1=1 ");
        Boolean allowFetchAll = businessModel.getAllowFetchAll();
        List<QueryField> queryFields = businessModel.getQueryFields();
        // return empty where part, aka query all data
        if (CollectionUtils.isEmpty(queryFields)) {
            return "";
        }
        queryFields.forEach(queryField -> {
            if (StringUtils.isNotBlank(queryField.getQueryExpression())) {
                String name = queryField.getName();
                // name must can't be blank
                if (StringUtils.isBlank(name)) {
                    throw new IllegalArgumentException("QueryField needs name attribute to match QueryExpression");
                }
                // append QueryExpression to whereBuilder
                whereBuilder.append(" and (").append(queryField.getQueryExpression()).append(") ");
            } else if (StringUtils.isNotBlank(queryField.getQueryOperator())) {
                Field field = queryField.getField();
                if (field == null || (field.getForQuery() != null && !field.getForQuery())) {
                    // field is not configured for Query
                    throw new IllegalArgumentException("field not configured proper");
                }
                String queryExpression = field.getQueryExpression();
                // if queryExpression is not blank, use it
                if (StringUtils.isNotBlank(queryExpression)) {
                    whereBuilder.append(" and (").append(queryExpression).append(") ");
                } else {
                    String physicalName = field.getPhysicalName();
                    String queryOperator = queryField.getQueryOperator();
                    String alias = businessModel.getAlias();
                    String name = field.getName();
                    whereBuilder.append(" and (");
                    if (StringUtils.isNotBlank(alias)) {
                        whereBuilder.append(alias).append(".");
                    }
                    whereBuilder.append(physicalName).append(" ").append(queryOperator).append(" :").append(name).append(")");
                }

            } else {
                // illegal if both of QueryExpression and QueryOperator is empty
                throw new IllegalArgumentException("Both of QueryExpression and QueryOperator is empty.");
            }
        });

        return whereBuilder.toString();
    }

    /**
     * parse query select part by using businessModel
     *
     * @return query select parse string
     */
    private String createSelect() {
        List<Field> fields = businessModel.getFields();
        if (CollectionUtils.isEmpty(fields)) {
            return "select * ";
        }
        StringBuilder selectBuilder = new StringBuilder(" select ");
        fields.forEach(field -> {
            String physicalName = field.getPhysicalName();
            Boolean forSelect = field.getForSelect();
            if (forSelect != null && !forSelect) {
                return;
            }
            String expression = field.getExpression();
            String name = field.getName();
            if (StringUtils.isNotBlank(expression)) {
                // use expression only
                selectBuilder.append(expression).append(" as ").append(name).append(",");
            } else if (StringUtils.isNotBlank(physicalName)) {
                // use physicalName
                selectBuilder.append(physicalName).append(" as ").append(name).append(",");
            }
        });
        // delete last comma(,)
        selectBuilder.deleteCharAt(selectBuilder.length() - 1);
        return selectBuilder.toString();
    }

    /**
     * parse query from part by using businessModel
     *
     * @return query from part String
     */
    private String createFrom() {
        StringBuilder fromBuilder = new StringBuilder(" from ");
        String baseTable = businessModel.getBaseTable();
        String alias = businessModel.getAlias();
        if (StringUtils.isBlank(baseTable)) {
            throw new IllegalArgumentException("baseTable can not be null");
        }
        fromBuilder.append(baseTable);
        if (StringUtils.isNotBlank(alias)) {
            fromBuilder.append(" ").append(alias);
        }
        return fromBuilder.toString();
    }

    public BusinessModel getBusinessModel() {
        return businessModel;
    }

    public void setBusinessModel(BusinessModel businessModel) {
        this.businessModel = businessModel;
    }
}
