package hls.core.websocket.components;

import com.hand.hap.message.MethodReflectUtils;
import com.hand.hap.message.websocket.WebSocketSessionManager;
import hls.core.sys.dto.SysUser;
import hls.core.sys.mapper.SysUserMapper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.SmartLifecycle;
import org.springframework.data.redis.connection.RedisConnection;
import org.springframework.data.redis.connection.RedisConnectionFactory;
import org.springframework.data.redis.serializer.RedisSerializer;
import org.springframework.scheduling.SchedulingAwareRunnable;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;
import redis.clients.jedis.Jedis;

import javax.annotation.PostConstruct;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * 以userId为基准,为notice初始化队列
 * <p>
 * Created by fjm on 2018/1/24.
 */
@Component
public class SysNoticeQueueInit implements DisposableBean, SmartLifecycle {

    private Logger logger = LoggerFactory.getLogger(SysNoticeQueueInit.class);

    private static final String QUEUE_HEADER = "hls:queue-";

    private Map<Long, SysNoticeQueueListener> listenerMap;

    private Map<Long, MonitorTask> monitorTaskMap;

    private static final int PHASE = 9999;

    private static final long MIN_RECOVERY_INTERVAL = 2000L;

    private static final long DEFAULT_RECOVERY_INTERVAL = 5000L;

    /**
     * 100ms.
     */
    private static final long IDLE_SLEEP_TIME = 100L;

    private long recoveryInterval = DEFAULT_RECOVERY_INTERVAL;

    private volatile boolean running = false;

    private ExecutorService executorService;

    @Autowired
    @Qualifier("v2redisConnectionFactory")
    private RedisConnectionFactory connectionFactory;

    @Autowired
    private RedisSerializer<String> stringRedisSerializer;

    @Autowired
    private SysUserMapper sysUserMapper;

    @Autowired
    private WebSocketSessionManager sessionManager;

    @Autowired
    private MessageBoxManager messageBoxManager;

    public long getRecoveryInterval() {
        return recoveryInterval;
    }

    public void setRecoveryInterval(long recoveryInterval) {
        this.recoveryInterval = recoveryInterval;
        if (recoveryInterval < MIN_RECOVERY_INTERVAL) {
            if (logger.isWarnEnabled()) {
                logger.warn("minimum for recoveryInterval is {}", MIN_RECOVERY_INTERVAL);
            }
            this.recoveryInterval = MIN_RECOVERY_INTERVAL;
        }
    }


    @PostConstruct
    public void contextInitialized() {
        if (listenerMap == null) {
            listenerMap = new HashMap<>();
        }
        if (monitorTaskMap == null) {
            monitorTaskMap = new HashMap<>();
        }
        List<SysUser> sysUsers = sysUserMapper.selectAll();
        executorService = Executors.newFixedThreadPool(sysUsers.size());//创建容量大小为用户数量的线程池
        sysUsers.forEach(item -> {
            Long userId = item.getUserId();
            String queue = QUEUE_HEADER + userId;
            Method[] methods = MessageBoxManager.class.getMethods();
            for (Method m : methods) {
                if ("onMessage".equalsIgnoreCase(m.getName())) {//方法名必须为onMessage,否则直接不处理
                    SysNoticeQueueListener listener = new SysNoticeQueueListener(queue, messageBoxManager, m);
                    listenerMap.put(userId, listener);
                    MonitorTask task = new MonitorTask(listener);
                    monitorTaskMap.put(userId, task);
                    break;
                }
            }
        });
    }

    @Override
    public void destroy() throws Exception {
        stop();
    }

    @Override
    public boolean isAutoStartup() {
        return true;
    }

    @Override
    public void stop(Runnable runnable) {
        stop();
        runnable.run();
    }

    @Override
    public void start() {
        if (!running) {
            running = true;

            if (logger.isDebugEnabled()) {
                logger.debug("startup success");
            }
        }
    }

    @Override
    public void stop() {
        if (isRunning()) {
            running = false;
            monitorTaskMap.forEach((k,v)->v.stop());
            executorService.shutdownNow();
            if (logger.isDebugEnabled()) {
                logger.debug("shutdown complete");
            }
        }
    }

    @Override
    public boolean isRunning() {
        return running;
    }

    @Override
    public int getPhase() {
        return PHASE;
    }


    /**
     * 内部类 用于实现消息的监听
     * 借鉴hap redis消息队列的实现
     */
    private static class SysNoticeQueueListener {
        private String queue;
        private Object target;
        private Method method;
        private RedisSerializer redisSerializer;
        private Logger logger;

        SysNoticeQueueListener(String queue, Object target, Method method) {
            this.queue = queue;
            this.target = target;
            this.method = method;
            this.redisSerializer = MethodReflectUtils.getProperRedisSerializer(method.getParameterTypes()[0]);
            this.logger = LoggerFactory.getLogger(target.getClass());
        }

        public String getQueue() {
            return queue;
        }

        public RedisSerializer getRedisSerializer() {
            return redisSerializer;
        }

        public void onQueueMessage(Object message, Long userId) {
            try {
                method.invoke(target, message, userId);
            } catch (Exception e) {
                Throwable thr = e;
                while (thr.getCause() != null) {
                    thr = thr.getCause();
                }
                if (logger.isErrorEnabled()) {
                    logger.error(thr.getMessage(), thr);
                }
            }
        }
    }


    /**
     * 内部类 用于创建消息队列的线程池
     * 借鉴hap redis消息队列的实现
     */
    public class MonitorTask implements SchedulingAwareRunnable {

        private SysNoticeQueueListener receiver;
        private RedisConnection connection;

        private boolean running = false;

        MonitorTask(SysNoticeQueueListener receiver) {
            this.receiver = receiver;
            Assert.notNull(receiver, "receiver is null.");
            Assert.hasText(receiver.getQueue(), "queue is not valid");
        }

        public void stop() {
            running = false;
            safeClose(true);
        }

        @Override
        public void run() {
            running = true;
            Object message;
            while (running) {
                String queue = receiver.getQueue();
                Long userId = Long.parseLong(queue.substring(QUEUE_HEADER.length()));//截取userId
                try {
                    if (connection == null) {
                        connection = connectionFactory.getConnection();
                    }
                    if (sessionManager.getSession(userId).isEmpty()) {//如果sessionMap中没有该userId对应的webscoketSession,则不消费该消息
                        sleep_(IDLE_SLEEP_TIME);
                        continue;
                    }
                    message = fetchMessage(connection, queue);
                    if (message == null) {
                        sleep_(IDLE_SLEEP_TIME);
                        continue;
                    }
                } catch (Throwable thr) {
                    if (!running) {
                        break;
                    }
                    safeClose();
                    if (logger.isDebugEnabled()) {
                        logger.error("exception occurred while get message from queue [" + queue + "]",
                                thr);
                        logger.debug("try recovery after {}ms", getRecoveryInterval());
                    }
                    sleep_(getRecoveryInterval());
                    continue;
                }
                try {
                    receiver.onQueueMessage(message, userId);
                } catch (Throwable thr) {
                    if (logger.isWarnEnabled()) {
                        logger.warn("exception occurred while receiver consume message.", thr);
                    }
                }
            }
            if (logger.isDebugEnabled()) {
                logger.debug("stop monitor:" + this);
            }
            safeClose();
        }

        Object fetchMessage(RedisConnection connection, String queue) {
            byte[] bytes = connection.rPop(stringRedisSerializer.serialize(queue));
            if (bytes == null || bytes.length == 0) {
                return null;
            }
            return receiver.getRedisSerializer().deserialize(bytes);
        }

        void safeClose(boolean... closeNative) {
            if (connection != null) {
                try {
                    if (closeNative.length > 0 && closeNative[0]) {
                        // close native connection to interrupt blocked
                        // operation
                        ((Jedis) connection.getNativeConnection()).disconnect();
                    }
                    connection.close();
                } catch (Exception e) {
                    // if (logger.isErrorEnabled()) {
                    // logger.error(e.getMessage(), e);
                    // }
                }
            }
            connection = null;
        }

        void sleep_(long time) {
            try {
                Thread.sleep(time);
            } catch (Exception e) {
                // if (logger.isErrorEnabled()) {
                // logger.error(e.getMessage(), e);
                // }
            }
        }

        @Override
        public boolean isLongLived() {
            return true;
        }
    }

    /**
     * 获取当前线程池
     *
     * @return
     */
    public ExecutorService getExecutorService() {
        return this.executorService;
    }


    /**
     * 获取当前的任务集合
     * @return
     */
    public Map<Long,MonitorTask> getMonitorTaskMap(){
        return this.monitorTaskMap;
    }

    /**
     * 根据userId创建任务,并添加到对应的listenerMap和monitorTaskMap中
     *
     * @param userId
     */
    public MonitorTask createMonitorTask(Long userId){
        MonitorTask task = null;
        String queue = QUEUE_HEADER + userId;
        Method[] methods = MessageBoxManager.class.getMethods();
        for (Method m : methods) {
            if ("onMessage".equalsIgnoreCase(m.getName())) {//方法名必须为onMessage,否则直接不处理
                SysNoticeQueueListener listener = new SysNoticeQueueListener(queue, messageBoxManager, m);
                listenerMap.put(userId, listener);
                task = new MonitorTask(listener);
                monitorTaskMap.put(userId, task);
                break;
            }
        }
        return task;
    }
}

